from dataclasses import dataclass
from datetime import date
from typing import Dict, Union

from typing_extensions import Annotated, Literal

from mashumaro.codecs import BasicDecoder
from mashumaro.codecs.orjson import ORJSONDecoder
from mashumaro.config import ADD_DIALECT_SUPPORT, BaseConfig
from mashumaro.dialect import Dialect
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import Discriminator


class DefaultDialect(Dialect):
    serialization_strategy = {
        date: {
            "deserialize": date.fromisoformat,
            "serialize": date.isoformat,
        },
        dict: {
            "deserialize": lambda x: {k: v for k, v in x},
            "serialize": lambda x: [(k, v) for k, v in x.items()],
        },
    }


class MyDialect(Dialect):
    serialization_strategy = {
        date: {
            "deserialize": date.fromordinal,
            "serialize": date.toordinal,
        },
        dict: {
            "deserialize": dict,
            "serialize": dict,
        },
    }


@dataclass
class Variant1(DataClassORJSONMixin):
    x: date
    y: Dict[str, int]

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]
        dialect = DefaultDialect
        discriminator = Discriminator(field="type", include_subtypes=True)


@dataclass
class _Variant1:
    x: date
    y: Dict[str, int]

    class Config(BaseConfig):
        discriminator = Discriminator(field="type", include_subtypes=True)


@dataclass
class Variant1Subtype1(Variant1):
    type: Literal[1] = 1


@dataclass
class _Variant1Subtype1(_Variant1):
    type: Literal[1] = 1


@dataclass
class Variant2(DataClassORJSONMixin):
    x: date
    y: Dict[str, int]

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]
        dialect = DefaultDialect


@dataclass
class _Variant2:
    x: date
    y: Dict[str, int]

    class Config(BaseConfig):
        dialect = DefaultDialect


@dataclass
class Variant2Subtype1(Variant2):
    type: Literal[1] = 1


@dataclass
class _Variant2Subtype1(_Variant2):
    type: Literal[1] = 1


@dataclass
class Variant2Wrapper(DataClassORJSONMixin):
    x: Annotated[Variant2, Discriminator(field="type", include_subtypes=True)]

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class _Variant2Wrapper:
    x: Annotated[_Variant2, Discriminator(field="type", include_subtypes=True)]

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class Variant3(DataClassORJSONMixin):
    x: date
    y: Dict[str, int]

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]
        dialect = DefaultDialect


@dataclass
class _Variant3:
    x: date
    y: Dict[str, int]

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]
        dialect = DefaultDialect


@dataclass
class Variant3Subtype(Variant3):
    type: Literal[3] = 3


@dataclass
class _Variant3Subtype(_Variant3):
    type: Literal[3] = 3


@dataclass
class Variant4(DataClassORJSONMixin):
    x: date
    y: Dict[str, int]

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]
        dialect = DefaultDialect


@dataclass
class _Variant4:
    x: date
    y: Dict[str, int]

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]
        dialect = DefaultDialect


@dataclass
class Variant4Subtype(Variant4):
    type: Literal[4] = 4


@dataclass
class _Variant4Subtype(_Variant4):
    type: Literal[4] = 4


@dataclass
class Variant34Wrapper(DataClassORJSONMixin):
    x: Annotated[
        Union[Variant3, Variant4],
        Discriminator(field="type", include_subtypes=True),
    ]

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class _Variant34Wrapper:
    x: Annotated[
        Union[_Variant3, _Variant4],
        Discriminator(field="type", include_subtypes=True),
    ]

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


def test_passing_dialect_to_config_based_variant_subtypes():
    assert Variant1.from_dict(
        {"type": 1, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}
    ) == Variant1Subtype1(date(2023, 6, 3), {"1": 2, "3": 4})
    decoder1 = BasicDecoder(_Variant1, default_dialect=DefaultDialect)
    assert decoder1.decode(
        {"type": 1, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}
    ) == _Variant1Subtype1(date(2023, 6, 3), {"1": 2, "3": 4})

    assert Variant1.from_json(
        '{"type": 1, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}'
    ) == Variant1Subtype1(date(2023, 6, 3), {"1": 2, "3": 4})
    decoder2 = ORJSONDecoder(_Variant1, default_dialect=DefaultDialect)
    assert decoder2.decode(
        '{"type": 1, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}'
    ) == _Variant1Subtype1(date(2023, 6, 3), {"1": 2, "3": 4})

    assert Variant1.from_dict(
        {"type": 1, "x": 738674, "y": {"1": 2, "3": 4}}, dialect=MyDialect
    ) == Variant1Subtype1(date(2023, 6, 3), {"1": 2, "3": 4})
    decoder3 = BasicDecoder(_Variant1, default_dialect=MyDialect)
    assert decoder3.decode(
        {"type": 1, "x": 738674, "y": {"1": 2, "3": 4}}
    ) == _Variant1Subtype1(date(2023, 6, 3), {"1": 2, "3": 4})

    assert Variant1.from_json(
        '{"type": 1, "x": 738674, "y": {"1": 2, "3": 4}}', dialect=MyDialect
    ) == Variant1Subtype1(date(2023, 6, 3), {"1": 2, "3": 4})
    decoder4 = ORJSONDecoder(_Variant1, default_dialect=MyDialect)
    assert decoder4.decode(
        '{"type": 1, "x": 738674, "y": {"1": 2, "3": 4}}'
    ) == _Variant1Subtype1(date(2023, 6, 3), {"1": 2, "3": 4})

    @dataclass
    class Variant1Subtype2(Variant1):
        type: Literal[2] = 2

    @dataclass
    class _Variant1Subtype2(_Variant1):
        type: Literal[2] = 2

    assert Variant1.from_dict(
        {"type": 2, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}
    ) == Variant1Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})
    assert decoder1.decode(
        {"type": 2, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}
    ) == _Variant1Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})

    assert Variant1.from_json(
        '{"type": 2, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}'
    ) == Variant1Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})
    assert decoder2.decode(
        '{"type": 2, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}'
    ) == _Variant1Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})

    assert Variant1.from_dict(
        {"type": 2, "x": 738674, "y": {"1": 2, "3": 4}}, dialect=MyDialect
    ) == Variant1Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})
    assert decoder3.decode(
        {"type": 2, "x": 738674, "y": {"1": 2, "3": 4}}
    ) == _Variant1Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})

    assert Variant1.from_json(
        '{"type": 2, "x": 738674, "y": {"1": 2, "3": 4}}', dialect=MyDialect
    ) == Variant1Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})
    assert decoder4.decode(
        '{"type": 2, "x": 738674, "y": {"1": 2, "3": 4}}'
    ) == _Variant1Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})


def test_passing_dialect_to_annotation_based_variant_subtypes():
    assert Variant2Wrapper.from_dict(
        {"x": {"type": 1, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == Variant2Wrapper(Variant2Subtype1(date(2023, 6, 3), {"1": 2, "3": 4}))
    decoder1 = BasicDecoder(_Variant2Wrapper)
    assert decoder1.decode(
        {"x": {"type": 1, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == _Variant2Wrapper(
        _Variant2Subtype1(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant2Wrapper.from_json(
        '{"x": {"type": 1, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == Variant2Wrapper(Variant2Subtype1(date(2023, 6, 3), {"1": 2, "3": 4}))
    decoder2 = ORJSONDecoder(_Variant2Wrapper)
    assert decoder2.decode(
        '{"x": {"type": 1, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == _Variant2Wrapper(
        _Variant2Subtype1(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant2Wrapper.from_dict(
        {"x": {"type": 1, "x": 738674, "y": {"1": 2, "3": 4}}},
        dialect=MyDialect,
    ) == Variant2Wrapper(Variant2Subtype1(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it

    assert Variant2Wrapper.from_json(
        '{"x": {"type": 1, "x": 738674, "y": {"1": 2, "3": 4}}}',
        dialect=MyDialect,
    ) == Variant2Wrapper(Variant2Subtype1(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it

    @dataclass
    class Variant2Subtype2(Variant2):
        type: Literal[2] = 2

    @dataclass
    class _Variant2Subtype2(_Variant2):
        type: Literal[2] = 2

    assert Variant2Wrapper.from_dict(
        {"x": {"type": 2, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == Variant2Wrapper(Variant2Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    assert decoder1.decode(
        {"x": {"type": 2, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == _Variant2Wrapper(
        _Variant2Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant2Wrapper.from_json(
        '{"x": {"type": 2, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == Variant2Wrapper(Variant2Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    assert decoder2.decode(
        '{"x": {"type": 2, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == _Variant2Wrapper(
        _Variant2Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant2Wrapper.from_dict(
        {"x": {"type": 2, "x": 738674, "y": {"1": 2, "3": 4}}},
        dialect=MyDialect,
    ) == Variant2Wrapper(Variant2Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it

    assert Variant2Wrapper.from_json(
        '{"x": {"type": 2, "x": 738674, "y": {"1": 2, "3": 4}}}',
        dialect=MyDialect,
    ) == Variant2Wrapper(Variant2Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it


def test_passing_dialect_to_annotation_based_union_subtypes():
    assert Variant34Wrapper.from_dict(
        {"x": {"type": 3, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == Variant34Wrapper(Variant3Subtype(date(2023, 6, 3), {"1": 2, "3": 4}))
    decoder1 = BasicDecoder(_Variant34Wrapper)
    assert decoder1.decode(
        {"x": {"type": 3, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == _Variant34Wrapper(
        _Variant3Subtype(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant34Wrapper.from_dict(
        {"x": {"type": 4, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == Variant34Wrapper(Variant4Subtype(date(2023, 6, 3), {"1": 2, "3": 4}))
    assert decoder1.decode(
        {"x": {"type": 4, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == _Variant34Wrapper(
        _Variant4Subtype(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant34Wrapper.from_json(
        '{"x": {"type": 3, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == Variant34Wrapper(Variant3Subtype(date(2023, 6, 3), {"1": 2, "3": 4}))
    decoder2 = ORJSONDecoder(_Variant34Wrapper)
    assert decoder2.decode(
        '{"x": {"type": 3, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == _Variant34Wrapper(
        _Variant3Subtype(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant34Wrapper.from_json(
        '{"x": {"type": 4, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == Variant34Wrapper(Variant4Subtype(date(2023, 6, 3), {"1": 2, "3": 4}))
    assert decoder2.decode(
        '{"x": {"type": 4, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == _Variant34Wrapper(
        _Variant4Subtype(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant34Wrapper.from_dict(
        {"x": {"type": 3, "x": 738674, "y": {"1": 2, "3": 4}}},
        dialect=MyDialect,
    ) == Variant34Wrapper(Variant3Subtype(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it

    assert Variant34Wrapper.from_dict(
        {"x": {"type": 4, "x": 738674, "y": {"1": 2, "3": 4}}},
        dialect=MyDialect,
    ) == Variant34Wrapper(Variant4Subtype(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it

    assert Variant34Wrapper.from_json(
        '{"x": {"type": 3, "x": 738674, "y": {"1": 2, "3": 4}}}',
        dialect=MyDialect,
    ) == Variant34Wrapper(Variant3Subtype(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it

    assert Variant34Wrapper.from_json(
        '{"x": {"type": 4, "x": 738674, "y": {"1": 2, "3": 4}}}',
        dialect=MyDialect,
    ) == Variant34Wrapper(Variant4Subtype(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it

    @dataclass
    class Variant3Subtype2(Variant3):
        type: Literal[5] = 5

    @dataclass
    class _Variant3Subtype2(_Variant3):
        type: Literal[5] = 5

    @dataclass
    class Variant4Subtype2(Variant4):
        type: Literal[6] = 6

    @dataclass
    class _Variant4Subtype2(_Variant4):
        type: Literal[6] = 6

    assert Variant34Wrapper.from_dict(
        {"x": {"type": 5, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == Variant34Wrapper(Variant3Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    assert decoder1.decode(
        {"x": {"type": 5, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == _Variant34Wrapper(
        _Variant3Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant34Wrapper.from_dict(
        {"x": {"type": 6, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == Variant34Wrapper(Variant4Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    assert decoder1.decode(
        {"x": {"type": 6, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}
    ) == _Variant34Wrapper(
        _Variant4Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant34Wrapper.from_json(
        '{"x": {"type": 5, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == Variant34Wrapper(Variant3Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    assert decoder2.decode(
        '{"x": {"type": 5, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == _Variant34Wrapper(
        _Variant3Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant34Wrapper.from_json(
        '{"x": {"type": 6, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == Variant34Wrapper(Variant4Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    assert decoder2.decode(
        '{"x": {"type": 6, "x": "2023-06-03", "y": [["1", 2], ["3", 4]]}}'
    ) == _Variant34Wrapper(
        _Variant4Subtype2(date(2023, 6, 3), {"1": 2, "3": 4})
    )

    assert Variant34Wrapper.from_dict(
        {"x": {"type": 5, "x": 738674, "y": {"1": 2, "3": 4}}},
        dialect=MyDialect,
    ) == Variant34Wrapper(Variant3Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it

    assert Variant34Wrapper.from_dict(
        {"x": {"type": 6, "x": 738674, "y": {"1": 2, "3": 4}}},
        dialect=MyDialect,
    ) == Variant34Wrapper(Variant4Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it

    assert Variant34Wrapper.from_json(
        '{"x": {"type": 5, "x": 738674, "y": {"1": 2, "3": 4}}}',
        dialect=MyDialect,
    ) == Variant34Wrapper(Variant3Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it

    assert Variant34Wrapper.from_json(
        '{"x": {"type": 6, "x": 738674, "y": {"1": 2, "3": 4}}}',
        dialect=MyDialect,
    ) == Variant34Wrapper(Variant4Subtype2(date(2023, 6, 3), {"1": 2, "3": 4}))
    # "decode" doesn't accept a dialect, so we can't pass MyDialect into it



from __future__ import annotations

from dataclasses import dataclass
from typing import TypedDict

from typing_extensions import NotRequired

from mashumaro import DataClassDictMixin
from mashumaro.codecs.basic import decode, encode
from mashumaro.jsonschema import build_json_schema


class MyDict(TypedDict):
    a: str
    b: NotRequired[int]


@dataclass
class MyDataClass(DataClassDictMixin):
    data: MyDict


def test_typeddict_with_not_required_and_future_annotations():
    # test workaround for https://github.com/Fatal1ty/mashumaro/issues/292
    assert decode({"a": "test", "b": 42}, MyDict) == {"a": "test", "b": 42}
    assert decode({"a": "test", "b": "42"}, MyDict) == {"a": "test", "b": 42}
    assert decode({"a": "test"}, MyDict) == {"a": "test"}

    assert encode(MyDict(a="test"), MyDict) == {"a": "test"}
    assert encode({"a": "test"}, MyDict) == {"a": "test"}
    assert encode(MyDict(a="test", b=42), MyDict) == {"a": "test", "b": 42}

    assert MyDataClass(MyDict(a="test", b=42)).to_dict() == {
        "data": {"a": "test", "b": 42}
    }
    assert MyDataClass(MyDict(a="test")).to_dict() == {"data": {"a": "test"}}

    assert MyDataClass.from_dict(
        {"data": {"a": "test", "b": 42}}
    ) == MyDataClass(MyDict(a="test", b=42))
    assert MyDataClass.from_dict({"data": {"a": "test"}}) == MyDataClass(
        MyDict(a="test")
    )


def test_jsonschema_for_not_required_and_future_annotations():
    schema = build_json_schema(MyDict).to_dict()
    assert schema == {
        "type": "object",
        "properties": {
            "a": {"type": "string"},
            "b": {"type": "integer"},
        },
        "required": ["a"],
        "additionalProperties": False,
    }



from dataclasses import is_dataclass
from inspect import cleandoc
from typing import Optional

from mashumaro.jsonschema.models import Context, JSONSchema
from mashumaro.jsonschema.schema import Instance


class BasePlugin:
    def get_schema(
        self,
        instance: Instance,
        ctx: Context,
        schema: Optional[JSONSchema] = None,
    ) -> Optional[JSONSchema]:
        pass


class DocstringDescriptionPlugin(BasePlugin):
    def get_schema(
        self,
        instance: Instance,
        ctx: Context,
        schema: Optional[JSONSchema] = None,
    ) -> Optional[JSONSchema]:
        if schema and is_dataclass(instance.type) and instance.type.__doc__:
            schema.description = cleandoc(instance.type.__doc__)
        return None



from collections.abc import Callable
from typing import Any, Literal, Optional, Type, TypedDict, Union

from mashumaro.core.const import Sentinel
from mashumaro.dialect import Dialect
from mashumaro.types import Discriminator, SerializationStrategy

__all__ = [
    "BaseConfig",
    "TO_DICT_ADD_BY_ALIAS_FLAG",
    "TO_DICT_ADD_OMIT_NONE_FLAG",
    "ADD_DIALECT_SUPPORT",
    "ADD_SERIALIZATION_CONTEXT",
    "SerializationStrategyValueType",
]


TO_DICT_ADD_BY_ALIAS_FLAG = "TO_DICT_ADD_BY_ALIAS_FLAG"
TO_DICT_ADD_OMIT_NONE_FLAG = "TO_DICT_ADD_OMIT_NONE_FLAG"
ADD_DIALECT_SUPPORT = "ADD_DIALECT_SUPPORT"
ADD_SERIALIZATION_CONTEXT = "ADD_SERIALIZATION_CONTEXT"


CodeGenerationOption = Literal[
    "TO_DICT_ADD_BY_ALIAS_FLAG",
    "TO_DICT_ADD_OMIT_NONE_FLAG",
    "ADD_DIALECT_SUPPORT",
    "ADD_SERIALIZATION_CONTEXT",
]


class SerializationStrategyDict(TypedDict, total=False):
    serialize: Union[str, Callable]
    deserialize: Union[str, Callable]


SerializationStrategyValueType = Union[
    SerializationStrategy, SerializationStrategyDict
]


class BaseConfig:
    debug: bool = False
    code_generation_options: list[CodeGenerationOption] = []
    serialization_strategy: dict[Any, SerializationStrategyValueType] = {}
    aliases: dict[str, str] = {}
    serialize_by_alias: Union[bool, Literal[Sentinel.MISSING]] = (
        Sentinel.MISSING
    )
    namedtuple_as_dict: Union[bool, Literal[Sentinel.MISSING]] = (
        Sentinel.MISSING
    )
    allow_postponed_evaluation: bool = True
    dialect: Optional[Type[Dialect]] = None
    omit_none: Union[bool, Literal[Sentinel.MISSING]] = Sentinel.MISSING
    omit_default: Union[bool, Literal[Sentinel.MISSING]] = Sentinel.MISSING
    orjson_options: Optional[int] = 0
    json_schema: dict[str, Any] = {}
    discriminator: Optional[Discriminator] = None
    lazy_compilation: bool = False
    sort_keys: bool = False
    allow_deserialization_not_by_alias: bool = False
    forbid_extra_keys: bool = False



from dataclasses import dataclass
from typing import Optional

import orjson
from typing_extensions import Self

from mashumaro import DataClassDictMixin
from mashumaro.codecs import BasicDecoder, BasicEncoder
from mashumaro.codecs.orjson import ORJSONDecoder, ORJSONEncoder
from mashumaro.mixins.orjson import DataClassORJSONMixin


@dataclass
class DataClassDict(DataClassDictMixin):
    next: Optional[Self] = None


@dataclass
class DataClassDictChild(DataClassDict):
    x: int = 42


@dataclass
class DataClassDictWithoutMixin:
    next: Optional[Self] = None


@dataclass
class DataClassOrjson(DataClassORJSONMixin):
    next: Optional[Self] = None


@dataclass
class DataClassOrjsonChild(DataClassOrjson):
    x: int = 42


def test_dataclass_dict_with_self():
    obj = DataClassDict(DataClassDict())
    assert obj.to_dict() == {"next": {"next": None}}
    assert DataClassDict.from_dict({"next": {"next": None}}) == obj
    assert DataClassDict().to_dict() == {"next": None}
    assert DataClassDict.from_dict({"next": None}) == DataClassDict()
    assert DataClassDict.from_dict({}) == DataClassDict()


def test_dataclass_dict_with_self_without_mixin():
    decoder = BasicDecoder(DataClassDictWithoutMixin)
    encoder = BasicEncoder(DataClassDictWithoutMixin)
    obj = DataClassDictWithoutMixin(DataClassDictWithoutMixin())
    assert encoder.encode(obj) == {"next": {"next": None}}
    assert decoder.decode({"next": {"next": None}}) == obj
    assert encoder.encode(DataClassDictWithoutMixin()) == {"next": None}
    assert decoder.decode({"next": None}) == DataClassDictWithoutMixin()
    assert decoder.decode({}) == DataClassDictWithoutMixin()


def test_dataclass_dict_child_with_self():
    obj = DataClassDictChild(DataClassDictChild())
    assert obj.to_dict() == {"x": 42, "next": {"x": 42, "next": None}}
    assert DataClassDictChild.from_dict({"next": {"next": None}}) == obj
    assert DataClassDictChild().to_dict() == {"x": 42, "next": None}
    assert DataClassDictChild.from_dict({"next": None}) == DataClassDictChild()
    assert DataClassDictChild.from_dict({}) == DataClassDictChild()


def test_dataclass_orjson_with_self():
    obj = DataClassOrjson(DataClassOrjson())
    assert obj.to_dict() == {"next": {"next": None}}
    assert DataClassOrjson.from_dict({"next": {"next": None}}) == obj
    assert DataClassOrjson().to_dict() == {"next": None}
    assert DataClassOrjson.from_dict({"next": None}) == DataClassOrjson()
    assert DataClassOrjson.from_dict({}) == DataClassOrjson()

    dump = orjson.dumps({"next": {"next": None}})
    assert obj.to_jsonb() == dump
    assert DataClassOrjson.from_json(dump) == obj
    dump = orjson.dumps({"next": None})
    assert DataClassOrjson().to_jsonb() == dump
    assert DataClassOrjson.from_json(dump) == DataClassOrjson()
    assert DataClassOrjson.from_json(b"{}") == DataClassOrjson()


def test_dataclass_orjson_with_self_without_mixin():
    decoder = ORJSONDecoder(DataClassDictWithoutMixin)
    encoder = ORJSONEncoder(DataClassDictWithoutMixin)
    obj = DataClassDictWithoutMixin(DataClassDictWithoutMixin())
    dump = orjson.dumps({"next": {"next": None}})
    assert encoder.encode(obj) == dump
    assert decoder.decode(dump) == obj
    dump = orjson.dumps({"next": None})
    assert encoder.encode(DataClassDictWithoutMixin()) == dump
    assert decoder.decode(dump) == DataClassDictWithoutMixin()
    assert decoder.decode(b"{}") == DataClassDictWithoutMixin()


def test_dataclass_orjson_child_with_self():
    obj = DataClassOrjsonChild(DataClassOrjsonChild())
    assert obj.to_dict() == {"x": 42, "next": {"x": 42, "next": None}}
    assert DataClassOrjsonChild.from_dict({"next": {"next": None}}) == obj
    assert DataClassOrjsonChild().to_dict() == {"x": 42, "next": None}
    assert (
        DataClassOrjsonChild.from_dict({"next": None})
        == DataClassOrjsonChild()
    )
    assert DataClassOrjsonChild.from_dict({}) == DataClassOrjsonChild()

    dump = orjson.dumps({"next": {"next": None, "x": 42}, "x": 42})
    assert obj.to_jsonb() == dump
    assert DataClassOrjsonChild.from_json(dump) == obj
    dump = orjson.dumps({"next": None, "x": 42})
    assert DataClassOrjsonChild().to_jsonb() == dump
    assert DataClassOrjsonChild.from_json(dump) == DataClassOrjsonChild()
    assert DataClassOrjsonChild.from_json(b"{}") == DataClassOrjsonChild()



import collections
import datetime
import ipaddress
import os
from base64 import encodebytes
from dataclasses import dataclass, field
from decimal import Decimal
from fractions import Fraction
from pathlib import (
    Path,
    PosixPath,
    PurePath,
    PurePosixPath,
    PureWindowsPath,
    WindowsPath,
)
from typing import (
    AbstractSet,
    Any,
    ByteString,
    ChainMap,
    Counter,
    DefaultDict,
    Deque,
    Dict,
    FrozenSet,
    Generic,
    List,
    Mapping,
    MutableMapping,
    Optional,
    OrderedDict,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)
from uuid import UUID
from zoneinfo import ZoneInfo

import pytest
from typing_extensions import Annotated, Literal, TypeVarTuple, Unpack

from mashumaro.config import BaseConfig
from mashumaro.core.meta.helpers import type_name
from mashumaro.helper import pass_through
from mashumaro.jsonschema.annotations import (
    Contains,
    DependentRequired,
    ExclusiveMaximum,
    ExclusiveMinimum,
    MaxContains,
    Maximum,
    MaxItems,
    MaxLength,
    MaxProperties,
    MinContains,
    Minimum,
    MinItems,
    MinLength,
    MinProperties,
    MultipleOf,
    Pattern,
    UniqueItems,
)
from mashumaro.jsonschema.builder import JSONSchemaBuilder, build_json_schema
from mashumaro.jsonschema.dialects import DRAFT_2020_12, OPEN_API_3_1
from mashumaro.jsonschema.models import (
    Context,
    JSONArraySchema,
    JSONObjectSchema,
    JSONSchema,
    JSONSchemaInstanceFormat,
    JSONSchemaInstanceFormatExtension,
    JSONSchemaInstanceType,
    JSONSchemaStringFormat,
)
from mashumaro.jsonschema.plugins import BasePlugin
from mashumaro.jsonschema.schema import (
    UTC_OFFSET_PATTERN,
    EmptyJSONSchema,
    Instance,
)
from mashumaro.types import Discriminator, SerializationStrategy
from tests.entities import (
    CustomPath,
    GenericNamedTuple,
    GenericTypedDict,
    MyDatetimeNewType,
    MyEnum,
    MyFlag,
    MyIntEnum,
    MyIntFlag,
    MyNamedTuple,
    MyNamedTupleWithDefaults,
    MyNamedTupleWithOptional,
    MyNativeStrEnum,
    MyStrEnum,
    MyUntypedNamedTuple,
    MyUntypedNamedTupleWithDefaults,
    T,
    T_Optional_int,
    TAny,
    TInt,
    TIntStr,
    TypedDictOptionalKeys,
    TypedDictOptionalKeysWithOptional,
    TypedDictRequiredAndOptionalKeys,
    TypedDictRequiredKeys,
    TypedDictRequiredKeysWithOptional,
    TypedDictWithReadOnly,
)
from tests.test_pep_655 import (
    TypedDictCorrectNotRequired,
    TypedDictCorrectRequired,
)

Ts = TypeVarTuple("Ts")


def dummy_serialize_as_str(_: Any) -> str:
    return "dummy"  # pragma: no cover


class ThirdPartyType:
    pass


@dataclass
class DataClassWithThirdPartyType:
    a: ThirdPartyType
    b: Optional[ThirdPartyType]
    c: ThirdPartyType = ThirdPartyType()
    d: Optional[ThirdPartyType] = None

    class Config(BaseConfig):
        serialization_strategy = {
            ThirdPartyType: {
                "deserialize": ThirdPartyType,
                "serialize": dummy_serialize_as_str,
            }
        }


def test_jsonschema_for_dataclass():
    @dataclass
    class MyClass:
        a: int
        b: float = 3.14
        c: Optional[int] = field(default=None, metadata={"alias": "cc"})
        d: str = ""
        e: int = field(init=False)
        f: List[int] = field(
            default_factory=list, metadata={"description": "description for f"}
        )

        class Config:
            aliases = {"a": "aa", "d": "dd"}

    assert build_json_schema(MyClass) == JSONObjectSchema(
        title="MyClass",
        properties={
            "aa": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            "b": JSONSchema(type=JSONSchemaInstanceType.NUMBER, default=3.14),
            "cc": JSONSchema(
                anyOf=[
                    JSONSchema(type=JSONSchemaInstanceType.INTEGER),
                    JSONSchema(type=JSONSchemaInstanceType.NULL),
                ],
                default=None,
            ),
            "dd": JSONSchema(type=JSONSchemaInstanceType.STRING, default=""),
            "f": JSONArraySchema(
                items=JSONSchema(type=JSONSchemaInstanceType.INTEGER),
                description="description for f",
            ),
        },
        additionalProperties=False,
        required=["aa"],
    )
    assert build_json_schema(MyClass, all_refs=True) == JSONSchema(
        reference="#/$defs/test_jsonschema_for_dataclass__locals__MyClass",
        definitions={
            "test_jsonschema_for_dataclass__locals__MyClass": JSONObjectSchema(
                title="test_jsonschema_for_dataclass__locals__MyClass",
                properties={
                    "aa": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
                    "b": JSONSchema(
                        type=JSONSchemaInstanceType.NUMBER, default=3.14
                    ),
                    "cc": JSONSchema(
                        anyOf=[
                            JSONSchema(type=JSONSchemaInstanceType.INTEGER),
                            JSONSchema(type=JSONSchemaInstanceType.NULL),
                        ],
                        default=None,
                    ),
                    "dd": JSONSchema(
                        type=JSONSchemaInstanceType.STRING, default=""
                    ),
                    "f": JSONArraySchema(
                        items=JSONSchema(type=JSONSchemaInstanceType.INTEGER),
                        description="description for f",
                    ),
                },
                additionalProperties=False,
                required=["aa"],
            )
        },
    )


def test_jsonschema_for_any():
    assert build_json_schema(Any) == EmptyJSONSchema()


def test_jsonschema_for_literal():
    assert build_json_schema(Literal[1]) == JSONSchema(const=1)
    assert build_json_schema(Literal[1, 2]) == JSONSchema(enum=[1, 2])
    assert build_json_schema(Literal["x", "y"]) == JSONSchema(enum=["x", "y"])
    assert build_json_schema(Literal[True, False]) == JSONSchema(
        enum=[True, False]
    )
    assert build_json_schema(Literal[1, None]) == JSONSchema(enum=[1, None])
    assert build_json_schema(Literal[MyEnum.a, MyEnum.b]) == JSONSchema(
        enum=["letter a", "letter b"]
    )
    assert build_json_schema(Literal[b"x", b"y"]) == JSONSchema(
        enum=[encodebytes(b"x").decode(), encodebytes(b"y").decode()]
    )


def test_jsonschema_for_special_typing_primitives():
    assert build_json_schema(Union[int, str]) == JSONSchema(
        anyOf=[
            JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            JSONSchema(type=JSONSchemaInstanceType.STRING),
        ]
    )
    assert build_json_schema(TAny) == EmptyJSONSchema()
    assert build_json_schema(T) == EmptyJSONSchema()
    assert build_json_schema(TInt) == JSONSchema(
        type=JSONSchemaInstanceType.INTEGER
    )
    assert build_json_schema(TIntStr) == JSONSchema(
        anyOf=[
            JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            JSONSchema(type=JSONSchemaInstanceType.STRING),
        ]
    )
    assert build_json_schema(T_Optional_int) == JSONSchema(
        anyOf=[
            JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            JSONSchema(type=JSONSchemaInstanceType.NULL),
        ]
    )
    assert build_json_schema(MyDatetimeNewType) == JSONSchema(
        type=JSONSchemaInstanceType.STRING,
        format=JSONSchemaStringFormat.DATETIME,
    )


@pytest.mark.parametrize(
    ("number_type", "instance_type"),
    (
        (int, JSONSchemaInstanceType.INTEGER),
        (float, JSONSchemaInstanceType.NUMBER),
    ),
)
def test_jsonschema_for_number(number_type, instance_type):
    assert build_json_schema(number_type) == JSONSchema(type=instance_type)
    assert build_json_schema(
        Annotated[
            number_type,
            Minimum(1),
            Maximum(2),
            ExclusiveMinimum(0),
            ExclusiveMaximum(3),
            MultipleOf(2),
        ]
    ) == JSONSchema(
        type=instance_type,
        minimum=1,
        maximum=2,
        exclusiveMinimum=0,
        exclusiveMaximum=3,
        multipleOf=2,
    )


def test_jsonschema_for_bool():
    assert build_json_schema(bool) == JSONSchema(
        type=JSONSchemaInstanceType.BOOLEAN
    )


def test_jsonschema_for_none():
    for instance_type in (None, type(None)):
        assert build_json_schema(instance_type) == JSONSchema(
            type=JSONSchemaInstanceType.NULL
        )


@pytest.mark.parametrize(
    ("instance_type", "string_format"),
    (
        (datetime.datetime, JSONSchemaStringFormat.DATETIME),
        (datetime.date, JSONSchemaStringFormat.DATE),
        (datetime.time, JSONSchemaStringFormat.TIME),
    ),
)
def test_jsonschema_for_datetime_objects(instance_type, string_format):
    assert build_json_schema(instance_type) == JSONSchema(
        type=JSONSchemaInstanceType.STRING, format=string_format
    )


def test_jsonschema_for_timedelta():
    assert build_json_schema(datetime.timedelta) == JSONSchema(
        type=JSONSchemaInstanceType.NUMBER,
        format=JSONSchemaInstanceFormatExtension.TIMEDELTA,
    )


def test_jsonschema_for_timezone():
    assert build_json_schema(datetime.timezone) == JSONSchema(
        type=JSONSchemaInstanceType.STRING,
        pattern=UTC_OFFSET_PATTERN,
    )


def test_jsonschema_for_zone_info():
    assert build_json_schema(ZoneInfo) == JSONSchema(
        type=JSONSchemaInstanceType.STRING,
        format=JSONSchemaInstanceFormatExtension.TIME_ZONE,
    )


def test_jsonschema_for_uuid():
    assert build_json_schema(UUID) == JSONSchema(
        type=JSONSchemaInstanceType.STRING, format=JSONSchemaStringFormat.UUID
    )


@pytest.mark.parametrize(
    ("instance_type", "string_format"),
    (
        (ipaddress.IPv4Address, JSONSchemaStringFormat.IPV4ADDRESS),
        (ipaddress.IPv6Address, JSONSchemaStringFormat.IPV6ADDRESS),
        (ipaddress.IPv4Network, JSONSchemaInstanceFormatExtension.IPV4NETWORK),
        (ipaddress.IPv6Network, JSONSchemaInstanceFormatExtension.IPV6NETWORK),
        (ipaddress.IPv4Network, JSONSchemaInstanceFormatExtension.IPV4NETWORK),
        (ipaddress.IPv6Network, JSONSchemaInstanceFormatExtension.IPV6NETWORK),
    ),
)
def test_jsonschema_for_ipaddress(instance_type, string_format):
    assert build_json_schema(instance_type) == JSONSchema(
        type=JSONSchemaInstanceType.STRING,
        format=string_format,
    )


def test_jsonschema_for_decimal():
    assert build_json_schema(Decimal) == JSONSchema(
        type=JSONSchemaInstanceType.STRING,
        format=JSONSchemaInstanceFormatExtension.DECIMAL,
    )


def test_jsonschema_for_fraction():
    assert build_json_schema(Fraction) == JSONSchema(
        type=JSONSchemaInstanceType.STRING,
        format=JSONSchemaInstanceFormatExtension.FRACTION,
    )


def test_jsonschema_for_bytestring():
    for instance_type in (ByteString, bytes, bytearray):
        assert build_json_schema(instance_type) == JSONSchema(
            type=JSONSchemaInstanceType.STRING,
            format=JSONSchemaInstanceFormatExtension.BASE64,
        )


def test_jsonschema_for_str():
    assert build_json_schema(str) == JSONSchema(
        type=JSONSchemaInstanceType.STRING
    )
    assert build_json_schema(
        Annotated[str, MinLength(1), MaxLength(5), Pattern("$[a-z]+^")]
    ) == JSONSchema(
        type=JSONSchemaInstanceType.STRING,
        minLength=1,
        maxLength=5,
        pattern="$[a-z]+^",
    )


def test_jsonschema_for_list():
    assert build_json_schema(List[int]) == JSONArraySchema(
        items=JSONSchema(type=JSONSchemaInstanceType.INTEGER)
    )
    assert build_json_schema(List) == JSONArraySchema()
    assert build_json_schema(List[Any]) == JSONArraySchema()
    assert build_json_schema(Annotated[List, min])
    assert build_json_schema(list) == JSONArraySchema()


def test_jsonschema_for_deque():
    assert build_json_schema(Deque[int]) == JSONArraySchema(
        items=JSONSchema(type=JSONSchemaInstanceType.INTEGER),
    )
    assert build_json_schema(Deque) == JSONArraySchema()
    assert build_json_schema(Deque[Any]) == JSONArraySchema()
    assert build_json_schema(collections.deque) == JSONArraySchema()


def test_jsonschema_for_tuple():
    assert build_json_schema(Tuple[int]) == JSONArraySchema(
        prefixItems=[JSONSchema(type=JSONSchemaInstanceType.INTEGER)],
        maxItems=1,
        minItems=1,
    )
    assert build_json_schema(Tuple) == JSONArraySchema()
    assert build_json_schema(Tuple[()]) == JSONArraySchema(maxItems=0)
    assert build_json_schema(Tuple[Any]) == JSONArraySchema(
        prefixItems=[EmptyJSONSchema()], maxItems=1, minItems=1
    )
    assert build_json_schema(Tuple[Any, ...]) == JSONArraySchema()
    assert build_json_schema(Tuple[int, ...]) == JSONArraySchema(
        items=JSONSchema(type=JSONSchemaInstanceType.INTEGER)
    )
    assert build_json_schema(tuple) == JSONArraySchema()


def test_jsonschema_for_named_tuple():
    assert build_json_schema(MyNamedTuple) == JSONArraySchema(
        prefixItems=[
            JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            JSONSchema(type=JSONSchemaInstanceType.NUMBER),
        ],
        maxItems=2,
        minItems=2,
    )
    assert build_json_schema(MyNamedTupleWithDefaults) == JSONArraySchema(
        prefixItems=[
            JSONSchema(type=JSONSchemaInstanceType.INTEGER, default=1),
            JSONSchema(type=JSONSchemaInstanceType.NUMBER, default=2.0),
        ],
        maxItems=2,
        minItems=2,
    )
    assert build_json_schema(MyNamedTupleWithOptional) == JSONArraySchema(
        prefixItems=[
            JSONSchema(
                anyOf=[
                    JSONSchema(type=JSONSchemaInstanceType.INTEGER),
                    JSONSchema(type=JSONSchemaInstanceType.NULL),
                ]
            ),
            JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        ],
        maxItems=2,
        minItems=2,
    )
    assert build_json_schema(MyUntypedNamedTuple) == JSONArraySchema(
        prefixItems=[EmptyJSONSchema(), EmptyJSONSchema()],
        maxItems=2,
        minItems=2,
    )
    assert build_json_schema(
        MyUntypedNamedTupleWithDefaults
    ) == JSONArraySchema(
        prefixItems=[JSONSchema(default=1), JSONSchema(default=2.0)],
        maxItems=2,
        minItems=2,
    )
    assert build_json_schema(GenericNamedTuple) == JSONArraySchema(
        prefixItems=[
            EmptyJSONSchema(),
            JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        ],
        maxItems=2,
        minItems=2,
    )
    assert build_json_schema(GenericNamedTuple[float]) == JSONArraySchema(
        prefixItems=[
            JSONSchema(type=JSONSchemaInstanceType.NUMBER),
            JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        ],
        maxItems=2,
        minItems=2,
    )


def test_jsonschema_for_named_tuple_as_dict():
    @dataclass
    class DataClassA:
        x: MyNamedTuple = field(metadata={"serialize": "as_dict"})

    @dataclass
    class DataClassB:
        x: MyNamedTuple

        class Config:
            namedtuple_as_dict = True

    schema = JSONObjectSchema(
        properties={
            "i": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            "f": JSONSchema(type=JSONSchemaInstanceType.NUMBER),
        },
        additionalProperties=False,
        required=["i", "f"],
    )
    assert build_json_schema(DataClassA).properties["x"] == schema
    assert build_json_schema(DataClassB).properties["x"] == schema


def test_jsonschema_for_named_tuple_as_list():
    @dataclass
    class DataClassA:
        x: MyNamedTuple = field(metadata={"serialize": "as_list"})

    @dataclass
    class DataClassB:
        x: MyNamedTuple

        class Config:
            namedtuple_as_dict = False

    schema = JSONArraySchema(
        prefixItems=[
            JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            JSONSchema(type=JSONSchemaInstanceType.NUMBER),
        ],
        maxItems=2,
        minItems=2,
    )
    assert build_json_schema(DataClassA).properties["x"] == schema
    assert build_json_schema(DataClassB).properties["x"] == schema


def test_jsonschema_for_named_tuple_with_overridden_serialization_method():
    class MySerializationStrategy(SerializationStrategy):
        def serialize(self, value: Any) -> MyNamedTuple: ...
        def deserialize(self, value: Any) -> Any: ...

    class MyAnySerializationStrategy(SerializationStrategy):
        def serialize(self, value: Any) -> Any: ...
        def deserialize(self, value: Any) -> Any: ...

    @dataclass
    class DataClassA:
        x: MyNamedTuple

        class Config:
            serialization_strategy = {MyNamedTuple: {"serialize": lambda x: x}}

    @dataclass
    class DataClassB:
        x: MyNamedTuple

        class Config:
            serialization_strategy = {
                MyNamedTuple: MyAnySerializationStrategy()
            }

    @dataclass
    class DataClassC:
        x: MyNamedTuple

        class Config:
            serialization_strategy = {MyNamedTuple: pass_through}

    @dataclass
    class DataClassD:
        x: MyNamedTuple

        class Config:
            serialization_strategy = {MyNamedTuple: MySerializationStrategy()}

    schema = JSONArraySchema(
        prefixItems=[
            JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            JSONSchema(type=JSONSchemaInstanceType.NUMBER),
        ],
        maxItems=2,
        minItems=2,
    )
    assert build_json_schema(DataClassA).properties["x"] == EmptyJSONSchema()
    assert build_json_schema(DataClassB).properties["x"] == EmptyJSONSchema()
    assert build_json_schema(DataClassC).properties["x"] == schema
    assert build_json_schema(DataClassD).properties["x"] == schema


def test_jsonschema_for_set():
    for generic_type in (FrozenSet, AbstractSet):
        assert build_json_schema(generic_type[int]) == JSONArraySchema(
            items=JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            uniqueItems=True,
        )
        assert build_json_schema(generic_type) == JSONArraySchema(
            uniqueItems=True
        )
        assert build_json_schema(generic_type[Any]) == JSONArraySchema(
            uniqueItems=True
        )
    assert build_json_schema(frozenset) == JSONArraySchema(uniqueItems=True)
    assert build_json_schema(set) == JSONArraySchema(uniqueItems=True)


def test_jsonschema_for_chainmap():
    assert build_json_schema(ChainMap[str, int]) == JSONArraySchema(
        items=JSONObjectSchema(
            propertyNames=JSONSchema(type=JSONSchemaInstanceType.STRING),
            additionalProperties=JSONSchema(
                type=JSONSchemaInstanceType.INTEGER
            ),
        )
    )
    assert build_json_schema(ChainMap) == JSONArraySchema(
        items=JSONObjectSchema()
    )
    assert build_json_schema(ChainMap[Any, Any]) == JSONArraySchema(
        items=JSONObjectSchema()
    )


def test_jsonschema_for_counter():
    assert build_json_schema(Counter[str]) == JSONObjectSchema(
        additionalProperties=JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        propertyNames=JSONSchema(type=JSONSchemaInstanceType.STRING),
    )
    assert build_json_schema(Counter) == JSONObjectSchema(
        additionalProperties=JSONSchema(type=JSONSchemaInstanceType.INTEGER)
    )
    assert build_json_schema(Counter[Any]) == JSONObjectSchema(
        additionalProperties=JSONSchema(type=JSONSchemaInstanceType.INTEGER),
    )
    assert build_json_schema(
        Annotated[
            Counter,
            MinProperties(1),
            MaxProperties(5),
            DependentRequired({"a": {"b"}, "b": {"c", "d"}}),
        ]
    ) == JSONObjectSchema(
        additionalProperties=JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        minProperties=1,
        maxProperties=5,
        dependentRequired={"a": {"b"}, "b": {"c", "d"}},
    )


def test_jsonschema_for_typeddict():
    assert build_json_schema(TypedDictRequiredKeys) == JSONObjectSchema(
        properties={
            "int": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            "float": JSONSchema(type=JSONSchemaInstanceType.NUMBER),
        },
        additionalProperties=False,
        required=["float", "int"],
    )
    assert build_json_schema(TypedDictOptionalKeys) == JSONObjectSchema(
        properties={
            "int": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            "float": JSONSchema(type=JSONSchemaInstanceType.NUMBER),
        },
        additionalProperties=False,
    )
    assert build_json_schema(
        TypedDictRequiredAndOptionalKeys
    ) == JSONObjectSchema(
        properties={
            "int": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            "float": JSONSchema(type=JSONSchemaInstanceType.NUMBER),
            "str": JSONSchema(type=JSONSchemaInstanceType.STRING),
        },
        additionalProperties=False,
        required=["float", "int"],
    )
    assert build_json_schema(
        TypedDictRequiredKeysWithOptional
    ) == JSONObjectSchema(
        properties={
            "x": JSONSchema(
                anyOf=[
                    JSONSchema(type=JSONSchemaInstanceType.INTEGER),
                    JSONSchema(type=JSONSchemaInstanceType.NULL),
                ]
            ),
            "y": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        },
        additionalProperties=False,
        required=["x", "y"],
    )
    assert build_json_schema(
        TypedDictOptionalKeysWithOptional
    ) == JSONObjectSchema(
        properties={
            "x": JSONSchema(
                anyOf=[
                    JSONSchema(type=JSONSchemaInstanceType.INTEGER),
                    JSONSchema(type=JSONSchemaInstanceType.NULL),
                ]
            ),
            "y": JSONSchema(type=JSONSchemaInstanceType.NUMBER),
        },
        additionalProperties=False,
    )
    assert build_json_schema(GenericTypedDict) == JSONObjectSchema(
        properties={
            "x": EmptyJSONSchema(),
            "y": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        },
        additionalProperties=False,
        required=["x", "y"],
    )
    assert build_json_schema(GenericTypedDict[float]) == JSONObjectSchema(
        properties={
            "x": JSONSchema(type=JSONSchemaInstanceType.NUMBER),
            "y": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        },
        additionalProperties=False,
        required=["x", "y"],
    )
    assert build_json_schema(GenericTypedDict[Any]) == JSONObjectSchema(
        properties={
            "x": EmptyJSONSchema(),
            "y": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        },
        additionalProperties=False,
        required=["x", "y"],
    )
    assert build_json_schema(TypedDictCorrectNotRequired) == JSONObjectSchema(
        properties={
            "required": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            "not_required": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        },
        additionalProperties=False,
        required=["required"],
    )
    assert build_json_schema(TypedDictCorrectRequired) == JSONObjectSchema(
        properties={
            "required": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            "not_required": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        },
        additionalProperties=False,
        required=["required"],
    )
    assert build_json_schema(TypedDictWithReadOnly) == JSONObjectSchema(
        properties={"x": JSONSchema(type=JSONSchemaInstanceType.INTEGER)},
        additionalProperties=False,
        required=["x"],
    )


def test_jsonschema_for_mapping():
    for generic_type in (
        Dict,
        Mapping,
        MutableMapping,
        OrderedDict,
        DefaultDict,
    ):
        assert build_json_schema(generic_type[str, int]) == JSONObjectSchema(
            additionalProperties=JSONSchema(
                type=JSONSchemaInstanceType.INTEGER
            ),
            propertyNames=JSONSchema(type=JSONSchemaInstanceType.STRING),
        )
        assert build_json_schema(generic_type) == JSONObjectSchema()
        assert build_json_schema(generic_type[Any, Any]) == JSONObjectSchema()
        assert build_json_schema(
            Annotated[
                generic_type,
                MinProperties(1),
                MaxProperties(5),
                DependentRequired({"a": {"b"}, "b": {"c", "d"}}),
            ]
        ) == JSONObjectSchema(
            minProperties=1,
            maxProperties=5,
            dependentRequired={"a": {"b"}, "b": {"c", "d"}},
        )
    assert build_json_schema(
        DefaultDict[int, Dict[str, int]]
    ) == JSONObjectSchema(
        additionalProperties=JSONObjectSchema(
            additionalProperties=JSONSchema(
                type=JSONSchemaInstanceType.INTEGER,
            ),
            propertyNames=JSONSchema(type=JSONSchemaInstanceType.STRING),
        ),
        propertyNames=JSONSchema(type=JSONSchemaInstanceType.INTEGER),
    )
    assert build_json_schema(dict) == JSONObjectSchema()


def test_jsonschema_for_sequence():
    assert build_json_schema(Sequence[int]) == JSONArraySchema(
        items=JSONSchema(type=JSONSchemaInstanceType.INTEGER)
    )
    assert build_json_schema(Sequence) == JSONArraySchema()
    assert build_json_schema(Sequence[Any]) == JSONArraySchema()


def test_jsonschema_for_pathlike():
    for pathlike_type in (
        Path,
        PurePath,
        PosixPath,
        PurePosixPath,
        WindowsPath,
        PureWindowsPath,
        os.PathLike,
        CustomPath,
    ):
        assert build_json_schema(pathlike_type) == JSONSchema(
            type=JSONSchemaInstanceType.STRING,
            format=JSONSchemaInstanceFormatExtension.PATH,
        )
    assert build_json_schema(
        Annotated[Path, MinLength(10), MaxLength(20)]
    ) == JSONSchema(
        type=JSONSchemaInstanceType.STRING,
        format=JSONSchemaInstanceFormatExtension.PATH,
        minLength=10,
        maxLength=20,
    )


def test_jsonschema_for_array_like_with_constraints():
    for array_type in (List, Deque, Tuple, FrozenSet, AbstractSet):
        schema = build_json_schema(
            Annotated[array_type, MinItems(1), MaxItems(5)]
        )
        assert schema.minItems == 1
        assert schema.maxItems == 5

        schema = build_json_schema(Annotated[array_type, UniqueItems(True)])
        assert schema.uniqueItems

        schema = build_json_schema(Annotated[array_type, UniqueItems(False)])
        assert not schema.uniqueItems

        contains_schema = JSONSchema(type=JSONSchemaInstanceType.INTEGER)
        schema = build_json_schema(
            Annotated[array_type, Contains(contains_schema)]
        )
        assert schema.contains == contains_schema
        assert schema.minContains is None
        assert schema.maxContains is None

        schema = build_json_schema(
            Annotated[array_type, MinContains(1), MaxContains(2)]
        )
        assert schema.contains is None
        assert schema.minContains is None
        assert schema.maxContains is None

        schema = build_json_schema(
            Annotated[
                array_type,
                Contains(contains_schema),
                MinContains(1),
                MaxContains(2),
            ]
        )
        assert schema.contains == contains_schema
        assert schema.minContains == 1
        assert schema.maxContains == 2


def test_jsonschema_for_enum():
    assert build_json_schema(MyEnum) == JSONSchema(
        enum=["letter a", "letter b"]
    )
    assert build_json_schema(MyStrEnum) == JSONSchema(
        enum=["letter a", "letter b"]
    )
    assert build_json_schema(MyNativeStrEnum) == JSONSchema(
        enum=["letter a", "letter b"]
    )
    assert build_json_schema(MyIntEnum) == JSONSchema(enum=[1, 2])
    assert build_json_schema(MyFlag) == JSONSchema(enum=[1, 2])
    assert build_json_schema(MyIntFlag) == JSONSchema(enum=[1, 2])


def test_jsonschema_for_type_var_tuple():
    assert build_json_schema(Ts) == JSONArraySchema()
    assert build_json_schema(
        Tuple[Unpack[Tuple[float, str]]]
    ) == JSONArraySchema(
        prefixItems=[
            JSONSchema(type=JSONSchemaInstanceType.NUMBER),
            JSONSchema(type=JSONSchemaInstanceType.STRING),
        ],
        maxItems=2,
        minItems=2,
    )
    assert build_json_schema(
        Tuple[Tuple[Unpack[Tuple[float, ...]], str], int]
    ) == JSONArraySchema(
        prefixItems=[
            JSONArraySchema(minItems=1),
            JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        ],
        maxItems=2,
        minItems=2,
    )

    @dataclass
    class GenericDataClass(Generic[Unpack[Ts]]):
        x: Tuple[Unpack[Ts]]

    assert build_json_schema(
        GenericDataClass[Unpack[Tuple[int, float]], datetime.time]
    ) == JSONObjectSchema(
        title="GenericDataClass",
        properties={
            "x": JSONArraySchema(
                prefixItems=[
                    JSONSchema(type=JSONSchemaInstanceType.INTEGER),
                    JSONSchema(type=JSONSchemaInstanceType.NUMBER),
                    JSONSchema(
                        type=JSONSchemaInstanceType.STRING,
                        format=JSONSchemaStringFormat.TIME,
                    ),
                ],
                maxItems=3,
                minItems=3,
            )
        },
        additionalProperties=False,
        required=["x"],
    )


def test_jsonschema_for_unsupported_type():
    with pytest.raises(NotImplementedError):
        build_json_schema(object)


def test_overridden_serialization_method_without_signature():
    @dataclass
    class DataClass:
        x: datetime.datetime
        y: datetime.datetime = field(
            metadata={"serialize": datetime.datetime.timestamp}
        )

        class Config(BaseConfig):
            serialization_strategy = {
                datetime.datetime: {
                    "serialize": datetime.datetime.timestamp,
                }
            }

    with pytest.warns(
        UserWarning, match=f"Type Any will be used for {type_name(DataClass)}"
    ):
        assert (
            build_json_schema(DataClass).properties["x"] == EmptyJSONSchema()
        )
        assert (
            build_json_schema(DataClass).properties["y"] == EmptyJSONSchema()
        )


def test_overridden_serialization_method_without_return_annotation():
    def as_timestamp(dt: datetime.datetime):  # pragma: no cover
        return dt.timestamp()

    @dataclass
    class DataClass:
        x: datetime.datetime
        y: datetime.datetime = field(metadata={"serialize": as_timestamp})

        class Config(BaseConfig):
            serialization_strategy = {
                datetime.datetime: {"serialize": as_timestamp}
            }

    assert build_json_schema(DataClass).properties["x"] == EmptyJSONSchema()
    assert build_json_schema(DataClass).properties["y"] == EmptyJSONSchema()


def test_overridden_serialization_method_with_return_annotation():
    def as_timestamp(dt: datetime.datetime) -> float:
        return dt.timestamp()  # pragma: no cover

    def first_datetime_as_timestamp(
        seq: List[datetime.datetime],
    ) -> float:
        return as_timestamp(seq[0])  # pragma: no cover

    @dataclass
    class DataClass:
        a: datetime.datetime
        b: datetime.datetime = field(metadata={"serialize": as_timestamp})
        c: List[datetime.datetime]
        d: List[datetime.datetime] = field(
            metadata={"serialize": first_datetime_as_timestamp}
        )
        e: Optional[datetime.datetime]
        f: List[Optional[datetime.datetime]]

        class Config(BaseConfig):
            serialization_strategy = {
                datetime.datetime: {"serialize": as_timestamp}
            }

    schema = build_json_schema(DataClass)
    assert schema.properties["a"] == JSONSchema(
        type=JSONSchemaInstanceType.NUMBER
    )
    assert schema.properties["b"] == JSONSchema(
        type=JSONSchemaInstanceType.NUMBER
    )
    assert schema.properties["c"] == JSONArraySchema(
        items=JSONSchema(type=JSONSchemaInstanceType.NUMBER)
    )
    assert schema.properties["d"] == JSONSchema(
        type=JSONSchemaInstanceType.NUMBER
    )
    assert schema.properties["e"] == JSONSchema(
        anyOf=[
            JSONSchema(type=JSONSchemaInstanceType.NUMBER),
            JSONSchema(type=JSONSchemaInstanceType.NULL),
        ]
    )
    assert schema.properties["f"] == JSONArraySchema(
        items=JSONSchema(
            anyOf=[
                JSONSchema(type=JSONSchemaInstanceType.NUMBER),
                JSONSchema(type=JSONSchemaInstanceType.NULL),
            ]
        )
    )


@pytest.mark.parametrize(
    ("basic_type", "schema_type"),
    (
        (str, JSONSchemaInstanceType.STRING),
        (int, JSONSchemaInstanceType.INTEGER),
        (float, JSONSchemaInstanceType.NUMBER),
        (bool, JSONSchemaInstanceType.BOOLEAN),
    ),
)
def test_basic_type_as_overridden_serialization_method(
    basic_type, schema_type
):
    @dataclass
    class DataClass:
        x: ThirdPartyType
        y: List[ThirdPartyType]

        class Config(BaseConfig):
            serialization_strategy = {
                ThirdPartyType: {"serialize": basic_type}
            }

    assert build_json_schema(DataClass).properties["x"] == JSONSchema(
        type=schema_type
    )
    assert build_json_schema(DataClass).properties["y"] == JSONArraySchema(
        items=JSONSchema(type=schema_type)
    )


def test_dataclass_overridden_serialization_method():
    def serialize_as_str(value: Any) -> str:
        return str(value)  # pragma: no cover

    @dataclass
    class Inner:
        x: int

    @dataclass
    class DataClass:
        a: Inner
        b: Optional[Inner]
        c: List[Inner]
        d: List[Optional[Inner]]

        class Config(BaseConfig):
            serialization_strategy = {Inner: {"serialize": serialize_as_str}}

    schema = build_json_schema(DataClass)
    assert schema.properties["a"] == JSONSchema(
        type=JSONSchemaInstanceType.STRING
    )
    assert schema.properties["b"] == JSONSchema(
        anyOf=[
            JSONSchema(type=JSONSchemaInstanceType.STRING),
            JSONSchema(type=JSONSchemaInstanceType.NULL),
        ]
    )
    assert schema.properties["c"] == JSONArraySchema(
        items=JSONSchema(type=JSONSchemaInstanceType.STRING)
    )
    assert schema.properties["d"] == JSONArraySchema(
        items=JSONSchema(
            anyOf=[
                JSONSchema(type=JSONSchemaInstanceType.STRING),
                JSONSchema(type=JSONSchemaInstanceType.NULL),
            ]
        )
    )


def test_third_party_overridden_serialization_method():
    schema = build_json_schema(DataClassWithThirdPartyType)
    assert schema.properties["a"] == JSONSchema(
        type=JSONSchemaInstanceType.STRING
    )
    assert schema.properties["b"] == JSONSchema(
        anyOf=[
            JSONSchema(type=JSONSchemaInstanceType.STRING),
            JSONSchema(type=JSONSchemaInstanceType.NULL),
        ]
    )
    assert schema.properties["c"] == JSONSchema(
        type=JSONSchemaInstanceType.STRING, default="dummy"
    )
    assert schema.properties["d"] == JSONSchema(
        anyOf=[
            JSONSchema(type=JSONSchemaInstanceType.STRING),
            JSONSchema(type=JSONSchemaInstanceType.NULL),
        ],
        default=None,
    )


def test_jsonschema_with_override_for_properties():
    @dataclass
    class DataClass:
        x: str
        y: datetime.datetime

        class Config(BaseConfig):
            json_schema = {
                "properties": {
                    "x": {
                        "type": "string",
                        "format": "date-time",
                        "description": "Description for x",
                    },
                    "y": {
                        "type": "number",
                        "description": "Description for y",
                    },
                }
            }

    assert build_json_schema(DataClass).properties["x"] == JSONSchema(
        type=JSONSchemaInstanceType.STRING,
        format=JSONSchemaStringFormat.DATETIME,
        description="Description for x",
    )
    assert build_json_schema(DataClass).properties["y"] == JSONSchema(
        type=JSONSchemaInstanceType.NUMBER,
        description="Description for y",
    )


def test_jsonschema_with_dialect_uri():
    schema = build_json_schema(str, with_dialect_uri=True)
    assert schema.schema == DRAFT_2020_12.uri
    assert schema.to_dict()["$schema"] == DRAFT_2020_12.uri
    schema = build_json_schema(
        str, dialect=OPEN_API_3_1, with_dialect_uri=True
    )
    assert schema.schema == OPEN_API_3_1.uri
    assert schema.to_dict()["$schema"] == OPEN_API_3_1.uri


def test_jsonschema_with_ref_prefix():
    @dataclass
    class DataClass:
        pass

    schema = {
        "$ref": (
            "#/components/responses/"
            "test_jsonschema_with_ref_prefix__locals__DataClass"
        )
    }
    assert (
        build_json_schema(
            List[DataClass], all_refs=True, ref_prefix="#/components/responses"
        ).items.to_dict()
        == schema
    )
    assert (
        build_json_schema(
            List[DataClass],
            all_refs=True,
            ref_prefix="#/components/responses/",
        ).items.to_dict()
        == schema
    )
    assert (
        build_json_schema(
            List[DataClass],
            dialect=OPEN_API_3_1,
            ref_prefix="#/components/responses/",
        ).items.to_dict()
        == schema
    )

    assert (
        JSONSchemaBuilder(all_refs=True, ref_prefix="#/components/responses")
        .build(List[DataClass])
        .items.to_dict()
        == schema
    )
    assert (
        JSONSchemaBuilder(
            dialect=OPEN_API_3_1, ref_prefix="#/components/responses"
        )
        .build(List[DataClass])
        .items.to_dict()
        == schema
    )


def test_jsonschema_with_additional_properties_true():
    @dataclass
    class DataClass:
        x: int

        class Config(BaseConfig):
            json_schema = {"additionalProperties": True}

    schema = JSONObjectSchema(
        title="DataClass",
        properties={
            "x": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        },
        additionalProperties=True,
        required=["x"],
    )
    assert build_json_schema(DataClass) == schema


def test_jsonschema_with_additional_properties_schema():
    @dataclass
    class DataClass:
        x: int

        class Config(BaseConfig):
            json_schema = {
                "additionalProperties": JSONSchema(
                    type=JSONSchemaInstanceType.INTEGER
                )
            }

    schema = JSONObjectSchema(
        title="DataClass",
        properties={
            "x": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        },
        additionalProperties=JSONSchema(type=JSONSchemaInstanceType.INTEGER),
        required=["x"],
    )
    assert build_json_schema(DataClass) == schema


def test_jsonschema_with_discriminator_for_local_types():
    @dataclass
    class A:
        value = "a"

    @dataclass
    class B:
        value = "b"

    @dataclass
    class Main:
        value: Annotated[
            Union[A, B], Discriminator(field="value", include_supertypes=True)
        ]

    schema = JSONObjectSchema(
        title="Main",
        properties={
            "value": JSONSchema(
                anyOf=[
                    JSONObjectSchema(
                        type=JSONSchemaInstanceType.OBJECT,
                        title="A",
                        additionalProperties=False,
                    ),
                    JSONObjectSchema(
                        type=JSONSchemaInstanceType.OBJECT,
                        title="B",
                        additionalProperties=False,
                    ),
                ],
            ),
        },
        additionalProperties=False,
        required=["value"],
    )
    assert build_json_schema(Main) == schema


def test_jsonschema_with_discriminator_with_default_for_local_types():
    @dataclass
    class A:
        value = "a"

    @dataclass
    class B:
        value = "b"

    @dataclass
    class Main:
        value: Annotated[
            Union[A, B, None],
            Discriminator(field="value", include_supertypes=True),
        ] = None

    schema = JSONObjectSchema(
        title="Main",
        properties={
            "value": JSONSchema(
                anyOf=[
                    JSONObjectSchema(
                        type=JSONSchemaInstanceType.OBJECT,
                        title="A",
                        additionalProperties=False,
                    ),
                    JSONObjectSchema(
                        type=JSONSchemaInstanceType.OBJECT,
                        title="B",
                        additionalProperties=False,
                    ),
                    JSONSchema(
                        type=JSONSchemaInstanceType.NULL,
                    ),
                ],
                default=None,
            ),
        },
        additionalProperties=False,
    )
    assert build_json_schema(Main) == schema


def test_jsonschema_with_optional_discriminator_and_default_for_local_types():
    @dataclass
    class A:
        value = "a"

    @dataclass
    class B:
        value = "b"

    @dataclass
    class Main:
        value: Optional[
            Annotated[
                Union[A, B],
                Discriminator(field="value", include_supertypes=True),
            ]
        ] = None

    schema = JSONObjectSchema(
        title="Main",
        properties={
            "value": JSONSchema(
                anyOf=[
                    JSONSchema(
                        anyOf=[
                            JSONObjectSchema(
                                type=JSONSchemaInstanceType.OBJECT,
                                title="A",
                                additionalProperties=False,
                            ),
                            JSONObjectSchema(
                                type=JSONSchemaInstanceType.OBJECT,
                                title="B",
                                additionalProperties=False,
                            ),
                        ]
                    ),
                    JSONSchema(
                        type=JSONSchemaInstanceType.NULL,
                    ),
                ],
                default=None,
            ),
        },
        additionalProperties=False,
    )
    assert build_json_schema(Main) == schema


def test_jsonschema_with_custom_instance_format():
    class CustomJSONSchemaInstanceFormatPlugin(BasePlugin):
        def get_schema(
            self,
            instance: Instance,
            ctx: Context,
            schema: Optional[JSONSchema] = None,
        ) -> Optional[JSONSchema]:
            for annotation in instance.annotations:
                if isinstance(annotation, JSONSchemaInstanceFormat):
                    schema.format = annotation
            return schema

    class Custom1InstanceFormat(JSONSchemaInstanceFormat):
        CUSTOM1 = "custom1"

    class CustomInstanceFormatBase(JSONSchemaInstanceFormat):
        pass

    class Custom2InstanceFormat(CustomInstanceFormatBase):
        CUSTOM2 = "custom2"

    type1 = Annotated[str, Custom1InstanceFormat.CUSTOM1]
    schema1 = build_json_schema(
        type1, plugins=[CustomJSONSchemaInstanceFormatPlugin()]
    )
    assert schema1.format is Custom1InstanceFormat.CUSTOM1
    assert schema1.to_dict()["format"] == "custom1"

    type2 = Annotated[int, Custom2InstanceFormat.CUSTOM2]
    schema2 = build_json_schema(
        type2, plugins=[CustomJSONSchemaInstanceFormatPlugin()]
    )
    assert schema2.format is Custom2InstanceFormat.CUSTOM2
    assert schema2.to_dict()["format"] == "custom2"

    assert (
        JSONSchema.from_dict({"format": "custom1"}).format
        is Custom1InstanceFormat.CUSTOM1
    )
    assert (
        JSONSchema.from_dict({"format": "custom2"}).format
        is Custom2InstanceFormat.CUSTOM2
    )

    @dataclass
    class MyClass:
        x: str
        y: str

        class Config(BaseConfig):
            json_schema = {
                "properties": {
                    "x": {"type": "string", "format": "custom1"},
                    "y": {"type": "string", "format": "custom2"},
                }
            }

    schema3 = build_json_schema(MyClass)
    assert schema3 == JSONObjectSchema(
        title="MyClass",
        properties={
            "x": JSONSchema(
                type=JSONSchemaInstanceType.STRING,
                format=Custom1InstanceFormat.CUSTOM1,
            ),
            "y": JSONSchema(
                type=JSONSchemaInstanceType.STRING,
                format=Custom2InstanceFormat.CUSTOM2,
            ),
        },
        required=["x", "y"],
        additionalProperties=False,
    )


def test_jsonschema_for_generic_dataclass():
    T = TypeVar("T")

    @dataclass
    class MyClass(Generic[T]):
        x: T
        y: list[T]

    assert build_json_schema(MyClass) == JSONObjectSchema(
        title="MyClass",
        properties={
            "x": EmptyJSONSchema(),
            "y": JSONArraySchema(),
        },
        additionalProperties=False,
        required=["x", "y"],
    )
    assert build_json_schema(MyClass[int]) == JSONObjectSchema(
        title="MyClass",
        properties={
            "x": JSONSchema(type=JSONSchemaInstanceType.INTEGER),
            "y": JSONArraySchema(
                items=JSONSchema(type=JSONSchemaInstanceType.INTEGER)
            ),
        },
        additionalProperties=False,
        required=["x", "y"],
    )

    @dataclass
    class MyClass2(Generic[T]):
        z: MyClass[T]

    assert build_json_schema(MyClass2) == JSONObjectSchema(
        title="MyClass2",
        properties={
            "z": JSONObjectSchema(
                title="MyClass",
                properties={
                    "x": EmptyJSONSchema(),
                    "y": JSONArraySchema(),
                },
                additionalProperties=False,
                required=["x", "y"],
            )
        },
        additionalProperties=False,
        required=["z"],
    )
    assert build_json_schema(MyClass2[str]) == JSONObjectSchema(
        title="MyClass2",
        properties={
            "z": JSONObjectSchema(
                title="MyClass",
                properties={
                    "x": JSONSchema(type=JSONSchemaInstanceType.STRING),
                    "y": JSONArraySchema(
                        items=JSONSchema(type=JSONSchemaInstanceType.STRING)
                    ),
                },
                additionalProperties=False,
                required=["x", "y"],
            )
        },
        additionalProperties=False,
        required=["z"],
    )



from dataclasses import dataclass, field
from datetime import date, datetime, time, timezone
from pathlib import Path
from typing import Any, Dict, Optional, Union

import ciso8601
import pytest

from mashumaro import DataClassDictMixin
from mashumaro.exceptions import (
    UnserializableField,
    UnsupportedDeserializationEngine,
    UnsupportedSerializationEngine,
)
from mashumaro.types import SerializationStrategy

from .entities import (
    MutableString,
    MyList,
    MyNamedTuple,
    MyNamedTupleWithDefaults,
    MyUntypedNamedTuple,
    MyUntypedNamedTupleWithDefaults,
    ThirdPartyType,
    TIntStr,
    TypedDictRequiredKeys,
)


def test_ciso8601_datetime_parser():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: datetime = field(metadata={"deserialize": "ciso8601"})

    should_be = DataClass(x=datetime(2021, 1, 2, 3, 4, 5, tzinfo=timezone.utc))
    instance = DataClass.from_dict({"x": "2021-01-02T03:04:05Z"})
    assert instance == should_be


def test_ciso8601_date_parser():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: date = field(metadata={"deserialize": "ciso8601"})

    should_be = DataClass(x=date(2021, 1, 2))
    instance = DataClass.from_dict({"x": "2021-01-02T03:04:05Z"})
    assert instance == should_be


def test_ciso8601_time_parser():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: time = field(metadata={"deserialize": "ciso8601"})

    should_be = DataClass(x=time(3, 4, 5))
    instance = DataClass.from_dict({"x": "2021-01-02T03:04:05Z"})
    assert instance == should_be


def test_pendulum_datetime_parser():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: datetime = field(metadata={"deserialize": "pendulum"})

    should_be = DataClass(x=datetime(2008, 12, 29, 7, tzinfo=timezone.utc))
    instance = DataClass.from_dict({"x": "2009-W01 07:00"})
    assert instance == should_be


def test_pendulum_date_parser():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: date = field(metadata={"deserialize": "pendulum"})

    should_be = DataClass(x=date(2008, 12, 29))
    instance = DataClass.from_dict({"x": "2009-W01"})
    assert instance == should_be


def test_pendulum_time_parser():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: time = field(metadata={"deserialize": "pendulum"})

    should_be = DataClass(x=time(3, 4, 5))
    instance = DataClass.from_dict({"x": "2009-W01 03:04:05"})
    assert instance == should_be


def test_unsupported_datetime_parser_engine():
    with pytest.raises(UnsupportedDeserializationEngine):

        @dataclass
        class DataClass(DataClassDictMixin):
            x: datetime = field(metadata={"deserialize": "unsupported"})


def test_global_function_datetime_parser():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: datetime = field(
            metadata={"deserialize": ciso8601.parse_datetime_as_naive}
        )

    should_be = DataClass(x=datetime(2021, 1, 2, 3, 4, 5))
    instance = DataClass.from_dict({"x": "2021-01-02T03:04:05+03:00"})
    assert instance == should_be


def test_local_function_datetime_parser():
    def parse_dt(s):
        return ciso8601.parse_datetime_as_naive(s)

    @dataclass
    class DataClass(DataClassDictMixin):
        x: datetime = field(metadata={"deserialize": parse_dt})

    should_be = DataClass(x=datetime(2021, 1, 2, 3, 4, 5))
    instance = DataClass.from_dict({"x": "2021-01-02T03:04:05+03:00"})
    assert instance == should_be


def test_class_method_datetime_parser():
    class DateTimeParser:
        @classmethod
        def parse_dt(cls, s: str) -> datetime:
            return datetime.fromisoformat(s)

    @dataclass
    class DataClass(DataClassDictMixin):
        x: datetime = field(metadata={"deserialize": DateTimeParser.parse_dt})

    should_be = DataClass(x=datetime(2021, 1, 2, 3, 4, 5))
    instance = DataClass.from_dict({"x": "2021-01-02T03:04:05"})
    assert instance == should_be


def test_class_instance_method_datetime_parser():
    class DateTimeParser:
        def __call__(self, s: str) -> datetime:
            return datetime.fromisoformat(s)

    @dataclass
    class DataClass(DataClassDictMixin):
        x: datetime = field(metadata={"deserialize": DateTimeParser()})

    should_be = DataClass(x=datetime(2021, 1, 2, 3, 4, 5))
    instance = DataClass.from_dict({"x": "2021-01-02T03:04:05"})
    assert instance == should_be


def test_callable_class_instance_datetime_parser():
    class CallableDateTimeParser:
        def __call__(self, s):
            return ciso8601.parse_datetime(s)

    @dataclass
    class DataClass(DataClassDictMixin):
        x: datetime = field(metadata={"deserialize": CallableDateTimeParser()})

    should_be = DataClass(x=datetime(2021, 1, 2, 3, 4, 5, tzinfo=timezone.utc))
    instance = DataClass.from_dict({"x": "2021-01-02T03:04:05Z"})
    assert instance == should_be


def test_lambda_datetime_parser():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: datetime = field(
            metadata={"deserialize": lambda s: ciso8601.parse_datetime(s)}
        )

    should_be = DataClass(x=datetime(2021, 1, 2, 3, 4, 5, tzinfo=timezone.utc))
    instance = DataClass.from_dict({"x": "2021-01-02T03:04:05Z"})
    assert instance == should_be


def test_derived_dataclass_metadata_deserialize_option():
    @dataclass
    class A:
        x: datetime = field(metadata={"deserialize": ciso8601.parse_datetime})

    @dataclass
    class B(A, DataClassDictMixin):
        y: datetime = field(metadata={"deserialize": ciso8601.parse_datetime})

    should_be = B(
        x=datetime(2021, 1, 2, 3, 4, 5, tzinfo=timezone.utc),
        y=datetime(2021, 1, 2, 3, 4, 5, tzinfo=timezone.utc),
    )
    instance = B.from_dict(
        {"x": "2021-01-02T03:04:05Z", "y": "2021-01-02T03:04:05Z"}
    )
    assert instance == should_be


def test_derived_dataclass_metadata_deserialize_option_removed():
    class MyClass:
        pass

    @dataclass
    class A:
        x: MyClass = field(
            metadata={"deserialize": MyClass, "serialize": lambda obj: obj.i}
        )

    with pytest.raises(UnserializableField):

        @dataclass
        class _(A, DataClassDictMixin):
            x: MyClass


def test_bytearray_overridden():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: bytearray = field(
            metadata={"deserialize": lambda s: s.upper().encode()}
        )

    should_be = DataClass(x=bytearray(b"ABC"))
    instance = DataClass.from_dict({"x": "abc"})
    assert instance == should_be


def test_path_like_overridden():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Path = field(
            metadata={"deserialize": lambda s: Path(str(s).upper())}
        )

    should_be = DataClass(x=Path("/ABC"))
    instance = DataClass.from_dict({"x": "/abc"})
    assert instance == should_be


def test_datetime_serialize_option():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: datetime = field(
            metadata={"serialize": lambda v: v.strftime("%Y-%m-%d %H:%M:%S")}
        )

    should_be = {"x": "2021-01-02 03:04:05"}
    instance = DataClass(x=datetime(2021, 1, 2, 3, 4, 5, tzinfo=timezone.utc))
    assert instance.to_dict() == should_be


def test_third_party_type_overridden():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: ThirdPartyType = field(
            metadata={
                "deserialize": lambda v: ThirdPartyType(v),
                "serialize": lambda v: v.value,
            }
        )

    should_be = DataClass(x=ThirdPartyType(123))
    instance = DataClass.from_dict({"x": 123})
    assert instance == should_be
    assert instance.to_dict() == {"x": 123}


def test_serializable_type_overridden():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MutableString = field(
            metadata={
                "deserialize": lambda s: MutableString(s.upper()),
                "serialize": lambda v: str(v).lower(),
            }
        )

    should_be = DataClass(x=MutableString("ABC"))
    instance = DataClass.from_dict({"x": "abc"})
    assert instance == should_be
    assert instance.to_dict() == {"x": "abc"}


def test_optional_overridden():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Optional[ThirdPartyType] = field(
            metadata={
                "deserialize": lambda v: ThirdPartyType(v),
                "serialize": lambda v: v.value,
            }
        )

    instance = DataClass.from_dict({"x": 123})
    assert instance
    assert instance.x.value == 123
    dct = instance.to_dict()
    assert dct["x"] == 123


def test_union_overridden():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Union[float, int] = field(
            metadata={
                "deserialize": lambda v: int(v),
                "serialize": lambda v: int(v),
            }
        )

    instance = DataClass.from_dict({"x": 1.0})
    assert instance == DataClass(x=1)
    assert instance.to_dict() == {"x": 1}
    for attr in dir(DataClass):
        assert not attr.startswith("__unpack_union")
        assert not attr.startswith("__unpack_union")


def test_type_var_overridden():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: TIntStr = field(
            metadata={
                "deserialize": lambda v: v * 2,
                "serialize": lambda v: v * 2,
            }
        )

    instance = DataClass.from_dict({"x": "a"})
    assert instance == DataClass(x="aa")
    assert instance.to_dict() == {"x": "aaaa"}
    for attr in dir(DataClass):
        assert not attr.startswith("__unpack_type_var")
        assert not attr.startswith("__pack_type_var")


def test_serialization_strategy():
    class TestSerializationStrategy(SerializationStrategy):
        def serialize(self, value):
            return [value]

        def deserialize(self, value):
            return value[0]

    @dataclass
    class DataClass(DataClassDictMixin):
        x: int = field(
            metadata={"serialization_strategy": TestSerializationStrategy()}
        )

    instance = DataClass(x=123)
    assert DataClass.from_dict({"x": [123]}) == instance
    assert instance.to_dict() == {"x": [123]}


def test_collection_derived_custom_class():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyList = field(
            metadata={"serialize": lambda v: v, "deserialize": lambda v: v}
        )

    instance = DataClass(x=[1, 2, 3])
    assert DataClass.from_dict({"x": [1, 2, 3]}) == instance
    assert instance.to_dict() == {"x": [1, 2, 3]}


def test_dataclass_with_typed_dict_overridden():
    def serialize_x(x: TypedDictRequiredKeys) -> Dict[str, Any]:
        return {"int": int(x["int"]), "float": float(x["float"])}

    def deserialize_x(x: Dict[str, Any]) -> TypedDictRequiredKeys:
        return TypedDictRequiredKeys(int=x["int"], float=x["float"])

    @dataclass
    class DataClass(DataClassDictMixin):
        x: TypedDictRequiredKeys = field(
            metadata={"serialize": serialize_x, "deserialize": deserialize_x}
        )

    obj = DataClass(x=TypedDictRequiredKeys(int=1, float=2.0))
    data = {"x": {"int": 1, "float": 2.0}}
    assert DataClass.from_dict(data) == obj
    assert obj.to_dict() == data


def test_named_tuple_as_dict_engine():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyNamedTuple = field(
            metadata={"serialize": "as_dict", "deserialize": "as_dict"}
        )

    obj = DataClass(x=MyNamedTuple(i=1, f=2.0))
    assert obj.to_dict() == {"x": {"i": 1, "f": 2.0}}
    assert DataClass.from_dict({"x": {"i": 1, "f": 2.0}}) == obj


def test_named_tuple_with_defaults_as_dict_engine():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyNamedTupleWithDefaults = field(
            metadata={"serialize": "as_dict", "deserialize": "as_dict"}
        )

    obj = DataClass(x=MyNamedTupleWithDefaults(i=1, f=2.0))
    assert obj.to_dict() == {"x": {"i": 1, "f": 2.0}}
    assert DataClass.from_dict({"x": {"i": 1, "f": 2.0}}) == obj


def test_untyped_named_tuple_as_dict_engine():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyUntypedNamedTuple = field(
            metadata={"serialize": "as_dict", "deserialize": "as_dict"}
        )

    obj = DataClass(x=MyUntypedNamedTuple(i=1, f=2.0))
    assert obj.to_dict() == {"x": {"i": 1, "f": 2.0}}
    assert DataClass.from_dict({"x": {"i": 1, "f": 2.0}}) == obj


def test_untyped_named_tuple_with_defaults_as_dict_engine():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyUntypedNamedTupleWithDefaults = field(
            metadata={"serialize": "as_dict", "deserialize": "as_dict"}
        )

    obj = DataClass(x=MyUntypedNamedTupleWithDefaults(i=1, f=2.0))
    assert obj.to_dict() == {"x": {"i": 1, "f": 2.0}}
    assert DataClass.from_dict({"x": {"i": 1, "f": 2.0}}) == obj


def test_unsupported_named_tuple_deserialization_engine():
    with pytest.raises(UnsupportedDeserializationEngine):

        @dataclass
        class DataClass(DataClassDictMixin):
            x: MyNamedTuple = field(metadata={"deserialize": "unsupported"})

    with pytest.raises(UnsupportedDeserializationEngine):

        @dataclass
        class DataClass(DataClassDictMixin):
            x: MyNamedTupleWithDefaults = field(
                metadata={"deserialize": "unsupported"}
            )

    with pytest.raises(UnsupportedDeserializationEngine):

        @dataclass
        class DataClass(DataClassDictMixin):
            x: MyUntypedNamedTuple = field(
                metadata={"deserialize": "unsupported"}
            )

    with pytest.raises(UnsupportedDeserializationEngine):

        @dataclass
        class DataClass(DataClassDictMixin):
            x: MyUntypedNamedTupleWithDefaults = field(
                metadata={"deserialize": "unsupported"}
            )


def test_unsupported_named_tuple_serialization_engine():
    with pytest.raises(UnsupportedSerializationEngine):

        @dataclass
        class DataClass(DataClassDictMixin):
            x: MyNamedTuple = field(metadata={"serialize": "unsupported"})

    with pytest.raises(UnsupportedSerializationEngine):

        @dataclass
        class DataClass(DataClassDictMixin):
            x: MyNamedTupleWithDefaults = field(
                metadata={"serialize": "unsupported"}
            )

    with pytest.raises(UnsupportedSerializationEngine):

        @dataclass
        class DataClass(DataClassDictMixin):
            x: MyUntypedNamedTuple = field(
                metadata={"serialize": "unsupported"}
            )

    with pytest.raises(UnsupportedSerializationEngine):

        @dataclass
        class DataClass(DataClassDictMixin):
            x: MyUntypedNamedTupleWithDefaults = field(
                metadata={"serialize": "unsupported"}
            )


def test_field_metadata_omit_engine():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int = field(metadata={"serialize": "omit"})

    obj = DataClass(x=42)
    assert obj.to_dict() == {}
    assert DataClass.from_dict({"x": 42}) == obj



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Optional

import pyperf
from dataclasses_json import DataClassJsonMixin, config

from benchmark.common import AbstractBenchmark

datetime_config = config(
    decoder=datetime.fromisoformat, encoder=datetime.isoformat
)
optional_datetime_config = config(
    decoder=lambda o: datetime.fromisoformat(o) if o is not None else None,
    encoder=lambda o: o.isoformat() if isinstance(o, datetime) else None,
)
enum_config = config(encoder=lambda o: o.value)
optional_enum_config = config(
    encoder=lambda o: o.value if isinstance(o, Enum) else None
)


class IssueState(Enum):
    OPEN = "open"
    CLOSED = "closed"


class MilestoneState(Enum):
    OPEN = "open"
    CLOSED = "closed"


class IssueStateReason(Enum):
    COMPLETED = "completed"
    REOPENED = "reopened"
    NOT_PLANNED = "not_planned"


class AuthorAssociation(Enum):
    COLLABORATOR = "COLLABORATOR"
    CONTRIBUTOR = "CONTRIBUTOR"
    FIRST_TIMER = "FIRST_TIMER"
    FIRST_TIME_CONTRIBUTOR = "FIRST_TIME_CONTRIBUTOR"
    MANNEQUIN = "MANNEQUIN"
    MEMBER = "MEMBER"
    NONE = "NONE"
    OWNER = "OWNER"


@dataclass(slots=True)
class User(DataClassJsonMixin):
    login: str
    id: int
    node_id: str
    avatar_url: str
    gravatar_id: Optional[str]
    url: str
    html_url: str
    followers_url: str
    following_url: str
    gists_url: str
    starred_url: str
    subscriptions_url: str
    organizations_url: str
    repos_url: str
    events_url: str
    received_events_url: str
    type: str
    site_admin: bool
    name: Optional[str] = None
    email: Optional[str] = None
    starred_at: Optional[datetime] = field(
        default=None, metadata=optional_datetime_config
    )


@dataclass(slots=True)
class IssueLabel(DataClassJsonMixin):
    id: int
    node_id: str
    url: str
    name: str
    description: Optional[str]
    color: Optional[str]
    default: bool


@dataclass(slots=True)
class Milestone(DataClassJsonMixin):
    url: str
    html_url: str
    labels_url: str
    id: int
    node_id: str
    number: int
    title: str
    description: Optional[str]
    creator: Optional[User]
    open_issues: int
    closed_issues: int
    created_at: datetime = field(metadata=datetime_config)
    updated_at: datetime = field(metadata=datetime_config)
    closed_at: Optional[datetime] = field(metadata=optional_datetime_config)
    due_on: Optional[datetime] = field(metadata=optional_datetime_config)
    state: MilestoneState = MilestoneState.OPEN


@dataclass(slots=True)
class Reactions(DataClassJsonMixin):
    url: str
    total_count: int
    plus_one: int = field(metadata=config(field_name="+1"))
    minus_one: int = field(metadata=config(field_name="-1"))
    laugh: int
    confused: int
    heart: int
    hooray: int
    eyes: int
    rocket: int


@dataclass(slots=True)
class Issue(DataClassJsonMixin):
    id: int
    node_id: str
    url: str
    repository_url: str
    labels_url: str
    comments_url: str
    events_url: str
    html_url: str
    number: int
    state: IssueState = field(metadata=enum_config)
    state_reason: Optional[IssueStateReason] = field(
        metadata=optional_enum_config
    )
    title: str
    body: Optional[str]
    user: Optional[User]
    labels: list[IssueLabel]
    assignee: Optional[User]
    assignees: Optional[list[User]]
    milestone: Optional[Milestone]
    locked: bool
    active_lock_reason: Optional[str]
    comments: int
    closed_at: Optional[datetime] = field(metadata=optional_datetime_config)
    created_at: datetime = field(metadata=datetime_config)
    updated_at: datetime = field(metadata=datetime_config)
    closed_by: Optional[User]
    author_association: AuthorAssociation = field(metadata=enum_config)
    draft: bool = False
    body_html: Optional[str] = None
    body_text: Optional[str] = None
    timeline_url: Optional[str] = None
    reactions: Optional[Reactions] = None


class Benchmark(AbstractBenchmark):
    LIBRARY = "dataclasses-json"

    def warmup(self, data) -> None:
        Issue.from_dict(data).to_dict()

    def run_loader(self, data) -> pyperf.Benchmark:
        return self._bench_loader_func(Issue.from_dict, data)

    def run_dumper(self, data) -> pyperf.Benchmark:
        obj = Issue.from_dict(data)
        return self._bench_dumper_func(obj.to_dict)



from dataclasses import dataclass
from datetime import date, datetime, time
from typing import Dict, List
from uuid import UUID, uuid4

import orjson
import pytest

from mashumaro import DataClassDictMixin
from mashumaro.config import ADD_DIALECT_SUPPORT, BaseConfig
from mashumaro.dialect import Dialect
from mashumaro.mixins.orjson import DataClassORJSONMixin

serialization_strategy = {
    datetime: {
        "serialize": lambda dt: dt.strftime("%Y/%m/%d/%H/%M/%S"),
        "deserialize": lambda s: datetime.strptime(s, "%Y/%m/%d/%H/%M/%S"),
    },
    date: {
        "serialize": date.toordinal,
        "deserialize": date.fromordinal,
    },
    time: {
        "serialize": lambda t: t.strftime("%H/%M/%S"),
        "deserialize": lambda s: datetime.strptime(s, "%H/%M/%S").time(),
    },
    UUID: {
        "serialize": lambda x: f"uuid:{x}",
        "deserialize": lambda s: UUID(s[5:]),
    },
}


class MyDialect(Dialect):
    serialization_strategy = serialization_strategy


@dataclass
class InnerDataClass(DataClassDictMixin):
    x: str


@dataclass
class DataClass(DataClassORJSONMixin):
    x: str
    inner: InnerDataClass


def test_data_class_orjson_mixin():
    assert DataClassORJSONMixin.from_json("") is None
    assert DataClassORJSONMixin().to_jsonb() is None


def test_to_orjson():
    @dataclass
    class DataClass(DataClassORJSONMixin):
        x: List[int]

    dumped = orjson.dumps({"x": [1, 2, 3]})
    assert DataClass([1, 2, 3]).to_jsonb() == dumped
    assert DataClass([1, 2, 3]).to_json() == dumped.decode()


def test_from_orjson():
    @dataclass
    class DataClass(DataClassORJSONMixin):
        x: List[int]

    dumped = orjson.dumps({"x": [1, 2, 3]})
    assert DataClass.from_json(dumped) == DataClass([1, 2, 3])


def test_orjson_with_serialization_strategy():
    @dataclass
    class DataClass(DataClassORJSONMixin):
        datetime: List[datetime]
        date: List[date]
        time: List[time]
        uuid: List[UUID]

        class Config(BaseConfig):
            serialization_strategy = serialization_strategy

    _datetime = datetime(2022, 10, 12, 12, 54, 30)
    _date = date(2022, 10, 12)
    _time = time(12, 54, 30)
    _uuid = uuid4()
    _datetime_dumped = _datetime.strftime("%Y/%m/%d/%H/%M/%S")
    _date_dumped = _date.toordinal()
    _time_dumped = _time.strftime("%H/%M/%S")
    _uuid_dumped = f"uuid:{_uuid}"
    instance = DataClass([_datetime], [_date], [_time], [_uuid])
    dict_dumped = {
        "datetime": [_datetime_dumped],
        "date": [_date_dumped],
        "time": [_time_dumped],
        "uuid": [_uuid_dumped],
    }
    orjson_dumped = orjson.dumps(
        {
            "datetime": [_datetime_dumped],
            "date": [_date_dumped],
            "time": [_time_dumped],
            "uuid": [_uuid_dumped],
        }
    )
    assert DataClass.from_dict(dict_dumped) == instance
    assert DataClass.from_json(orjson_dumped) == instance
    assert instance.to_dict() == dict_dumped
    assert instance.to_jsonb() == orjson_dumped


def test_orjson_with_dialect():
    @dataclass
    class DataClass(DataClassORJSONMixin):
        datetime: List[datetime]
        date: List[date]
        time: List[time]
        uuid: List[UUID]

        class Config(BaseConfig):
            dialect = MyDialect

    _datetime = datetime(2022, 10, 12, 12, 54, 30)
    _date = date(2022, 10, 12)
    _time = time(12, 54, 30)
    _uuid = uuid4()
    _datetime_dumped = _datetime.strftime("%Y/%m/%d/%H/%M/%S")
    _date_dumped = _date.toordinal()
    _time_dumped = _time.strftime("%H/%M/%S")
    _uuid_dumped = f"uuid:{_uuid}"
    instance = DataClass([_datetime], [_date], [_time], [_uuid])
    dict_dumped = {
        "datetime": [_datetime_dumped],
        "date": [_date_dumped],
        "time": [_time_dumped],
        "uuid": [_uuid_dumped],
    }
    orjson_dumped = orjson.dumps(
        {
            "datetime": [_datetime_dumped],
            "date": [_date_dumped],
            "time": [_time_dumped],
            "uuid": [_uuid_dumped],
        }
    )
    assert DataClass.from_dict(dict_dumped) == instance
    assert DataClass.from_json(orjson_dumped) == instance
    assert instance.to_dict() == dict_dumped
    assert instance.to_jsonb() == orjson_dumped


def test_orjson_with_dialect_support():
    @dataclass
    class DataClass(DataClassORJSONMixin):
        datetime: List[datetime]
        date: List[date]
        time: List[time]
        uuid: List[UUID]

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]

    _datetime = datetime(2022, 10, 12, 12, 54, 30)
    _date = date(2022, 10, 12)
    _time = time(12, 54, 30)
    _uuid = uuid4()
    _datetime_dumped = _datetime.strftime("%Y/%m/%d/%H/%M/%S")
    _date_dumped = _date.toordinal()
    _time_dumped = _time.strftime("%H/%M/%S")
    _uuid_dumped = f"uuid:{_uuid}"
    instance = DataClass([_datetime], [_date], [_time], [_uuid])
    dict_dumped = {
        "datetime": [_datetime.isoformat()],
        "date": [_date.isoformat()],
        "time": [_time.isoformat()],
        "uuid": [str(_uuid)],
    }
    dict_dumped_dialect = {
        "datetime": [_datetime_dumped],
        "date": [_date_dumped],
        "time": [_time_dumped],
        "uuid": [_uuid_dumped],
    }
    orjson_dumped = orjson.dumps(
        {
            "datetime": [_datetime],
            "date": [_date],
            "time": [_time],
            "uuid": [_uuid],
        }
    )
    orjson_dumped_dialect = orjson.dumps(
        {
            "datetime": [_datetime_dumped],
            "date": [_date_dumped],
            "time": [_time_dumped],
            "uuid": [_uuid_dumped],
        }
    )
    assert DataClass.from_dict(dict_dumped) == instance
    assert (
        DataClass.from_dict(dict_dumped_dialect, dialect=MyDialect) == instance
    )
    assert DataClass.from_json(orjson_dumped) == instance
    assert (
        DataClass.from_json(orjson_dumped_dialect, dialect=MyDialect)
        == instance
    )
    assert instance.to_dict() == dict_dumped
    assert instance.to_dict(dialect=MyDialect) == dict_dumped_dialect
    assert instance.to_jsonb() == orjson_dumped
    assert instance.to_jsonb(dialect=MyDialect) == orjson_dumped_dialect


def test_orjson_with_custom_encoder_and_decoder():
    def decoder(data) -> Dict[str, bytes]:
        def to_lower(d):
            result = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    result[k] = to_lower(v)
                else:
                    result[k] = v.lower()
            return result

        return to_lower(orjson.loads(data))

    def encoder(data: Dict[str, bytes], **_) -> bytes:
        def to_upper(d):
            result = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    result[k] = to_upper(v)
                else:
                    result[k] = v.upper()
            return result

        return orjson.dumps(to_upper(data))

    instance = DataClass("abc", InnerDataClass("def"))
    dumped = orjson.dumps({"x": "ABC", "inner": {"x": "DEF"}})
    assert instance.to_jsonb(encoder=encoder) == dumped
    assert DataClass.from_json(dumped, decoder=decoder) == instance


def test_to_orjson_with_non_str_keys():
    @dataclass
    class DataClass(DataClassORJSONMixin):
        x: Dict[int, int]

    instance = DataClass({1: 2})
    dumped = b'{"x":{"1":2}}'
    with pytest.raises(TypeError):
        instance.to_jsonb()
    assert instance.to_jsonb(orjson_options=orjson.OPT_NON_STR_KEYS) == dumped
    assert DataClass.from_json(dumped) == instance



import enum
import sys

__all__ = [
    "PY_39",
    "PY_310",
    "PY_310_MIN",
    "PY_311_MIN",
    "PY_312_MIN",
    "PY_313_MIN",
    "PY_314_MIN",
    "Sentinel",
]


PY_39 = sys.version_info.major == 3 and sys.version_info.minor == 9
PY_310 = sys.version_info.major == 3 and sys.version_info.minor == 10
PY_311 = sys.version_info.major == 3 and sys.version_info.minor == 11
PY_312 = sys.version_info.major == 3 and sys.version_info.minor == 12
PY_313 = sys.version_info.major == 3 and sys.version_info.minor == 13
PY_314_MIN = sys.version_info.major == 3 and sys.version_info.minor >= 14

PY_313_MIN = PY_313 or PY_314_MIN
PY_312_MIN = PY_312 or PY_313_MIN
PY_311_MIN = PY_311 or PY_312_MIN
PY_310_MIN = PY_310 or PY_311_MIN


class Sentinel(enum.Enum):
    MISSING = enum.auto()



import datetime
import enum
import ipaddress
import os
import re
import sys
import typing
import uuid
import zoneinfo
from base64 import encodebytes
from collections import ChainMap, Counter, OrderedDict, deque
from collections.abc import Callable, Collection, Mapping, Sequence, Set
from contextlib import suppress
from dataclasses import is_dataclass
from decimal import Decimal
from fractions import Fraction
from typing import Any, ForwardRef, Optional, Tuple, Union

import typing_extensions
from typing_extensions import NotRequired

from mashumaro.core.const import PY_311_MIN
from mashumaro.core.meta.code.lines import CodeLines
from mashumaro.core.meta.helpers import (
    get_args,
    get_class_that_defines_method,
    get_function_return_annotation,
    get_literal_values,
    get_type_origin,
    get_type_var_default,
    is_final,
    is_generic,
    is_literal,
    is_named_tuple,
    is_new_type,
    is_not_required,
    is_optional,
    is_readonly,
    is_required,
    is_self,
    is_special_typing_primitive,
    is_type_alias_type,
    is_type_var,
    is_type_var_any,
    is_type_var_tuple,
    is_typed_dict,
    is_union,
    is_unpack,
    not_none_type_arg,
    resolve_type_params,
    substitute_type_params,
    type_name,
    type_var_has_default,
)
from mashumaro.core.meta.types.common import (
    Expression,
    ExpressionWrapper,
    NoneType,
    Registry,
    ValueSpec,
    clean_id,
    ensure_generic_collection,
    ensure_generic_collection_subclass,
    ensure_generic_mapping,
    expr_or_maybe_none,
    random_hex,
)
from mashumaro.exceptions import (
    UnserializableDataError,
    UnserializableField,
    UnsupportedSerializationEngine,
)
from mashumaro.helper import pass_through
from mashumaro.types import (
    GenericSerializableType,
    SerializableType,
    SerializationStrategy,
)

if sys.version_info >= (3, 14):
    from typing import evaluate_forward_ref

    from annotationlib import get_annotations
else:
    from typing_extensions import evaluate_forward_ref, get_annotations

__all__ = ["PackerRegistry"]


PackerRegistry = Registry()
register = PackerRegistry.register


def _pack_with_annotated_serialization_strategy(
    spec: ValueSpec,
    strategy: SerializationStrategy,
) -> Expression:
    strategy_type = type(strategy)
    try:
        value_type: Union[type, Any] = get_function_return_annotation(
            strategy.serialize
        )
    except (KeyError, ValueError):
        value_type = Any
    if isinstance(value_type, ForwardRef):
        value_type = evaluate_forward_ref(value_type)
    value_type = substitute_type_params(
        value_type,  # type: ignore
        resolve_type_params(strategy_type, get_args(spec.type))[strategy_type],
    )
    overridden_fn = f"__{spec.field_ctx.name}_serialize_{random_hex()}"
    setattr(spec.attrs, overridden_fn, strategy.serialize)
    new_spec = spec.copy(
        type=value_type,
        expression=(
            f"{spec.self_attrs_name}.{overridden_fn}({spec.expression})"
        ),
    )
    field_metadata = new_spec.field_ctx.metadata
    if field_metadata.get("serialization_strategy") is strategy:
        new_spec.field_ctx.metadata = {
            k: v
            for k, v in field_metadata.items()
            if k != "serialization_strategy"
        }
    return PackerRegistry.get(
        spec.copy(
            type=value_type,
            expression=(
                f"{spec.self_attrs_name}.{overridden_fn}({spec.expression})"
            ),
        )
    )


def get_overridden_serialization_method(
    spec: ValueSpec,
) -> Optional[Union[Callable, str, ExpressionWrapper]]:
    serialize_option = spec.field_ctx.metadata.get("serialize")
    if serialize_option is not None:
        return serialize_option
    checking_types = [spec.type, spec.origin_type]
    if spec.annotated_type:
        checking_types.insert(0, spec.annotated_type)
    for typ in checking_types:
        for strategy in spec.builder.iter_serialization_strategies(
            spec.field_ctx.metadata, typ
        ):
            if strategy is pass_through:
                return pass_through
            elif isinstance(strategy, dict):
                serialize_option = strategy.get("serialize")
            elif isinstance(strategy, SerializationStrategy):
                if strategy.__use_annotations__ or is_generic(type(strategy)):
                    return ExpressionWrapper(
                        _pack_with_annotated_serialization_strategy(
                            spec=spec,
                            strategy=strategy,
                        )
                    )
                else:
                    serialize_option = strategy.serialize
            if serialize_option is not None:
                return serialize_option


@register
def pack_type_with_overridden_serialization(
    spec: ValueSpec,
) -> Optional[Expression]:
    serialization_method = get_overridden_serialization_method(spec)
    if serialization_method is pass_through:
        return spec.expression
    elif isinstance(serialization_method, ExpressionWrapper):
        return serialization_method.expression
    elif callable(serialization_method):
        overridden_fn = f"__{spec.field_ctx.name}_serialize_{random_hex()}"
        setattr(spec.attrs, overridden_fn, staticmethod(serialization_method))
        return f"{spec.self_attrs_name}.{overridden_fn}({spec.expression})"


def _pack_annotated_serializable_type(
    spec: ValueSpec,
) -> Optional[Expression]:
    try:
        # noinspection PyProtectedMember
        # noinspection PyUnresolvedReferences
        value_type = get_function_return_annotation(
            spec.origin_type._serialize
        )
    except (KeyError, ValueError):
        raise UnserializableField(
            field_name=spec.field_ctx.name,
            field_type=spec.type,
            holder_class=spec.builder.cls,
            msg="Method _serialize must have return annotation",
        ) from None
    if is_self(value_type):
        return f"{spec.expression}._serialize()"
    if isinstance(value_type, ForwardRef):
        value_type = evaluate_forward_ref(value_type)
    value_type = substitute_type_params(
        value_type,
        resolve_type_params(spec.origin_type, get_args(spec.type))[
            spec.origin_type
        ],
    )
    return PackerRegistry.get(
        spec.copy(
            type=value_type,
            expression=f"{spec.expression}._serialize()",
        )
    )


@register
def pack_serializable_type(spec: ValueSpec) -> Optional[Expression]:
    try:
        if not issubclass(spec.origin_type, SerializableType):
            return None
    except TypeError:
        return None
    if spec.origin_type.__use_annotations__:
        return _pack_annotated_serializable_type(spec)
    else:
        return f"{spec.expression}._serialize()"


@register
def pack_generic_serializable_type(spec: ValueSpec) -> Optional[Expression]:
    with suppress(TypeError):
        if issubclass(spec.origin_type, GenericSerializableType):
            type_args = get_args(spec.type)
            spec.builder.add_type_modules(*type_args)
            type_arg_names = ", ".join(list(map(type_name, type_args)))
            return f"{spec.expression}._serialize([{type_arg_names}])"


@register
def pack_dataclass(spec: ValueSpec) -> Optional[Expression]:
    if is_dataclass(spec.origin_type):
        type_args = get_args(spec.type)
        method_name = spec.builder.get_pack_method_name(
            type_args, spec.builder.format_name
        )
        method_loc = spec.origin_type if spec.builder.is_nailed else spec.attrs
        if get_class_that_defines_method(
            method_name, method_loc
        ) != method_loc and (
            spec.origin_type is not spec.builder.cls
            or spec.builder.get_pack_method_name(
                type_args=type_args,
                format_name=spec.builder.format_name,
                encoder=spec.builder.encoder,
            )
            != method_name
        ):
            builder = spec.builder.__class__(
                spec.origin_type,
                type_args,
                dialect=spec.builder.dialect,
                format_name=spec.builder.format_name,
                default_dialect=spec.builder.default_dialect,
                attrs=method_loc,
                attrs_registry=(
                    spec.attrs_registry if not spec.builder.is_nailed else None
                ),
            )
            builder.add_pack_method()
        flags = spec.builder.get_pack_method_flags(spec.type)
        if spec.builder.is_nailed:
            return f"{spec.expression}.{method_name}({flags})"
        else:
            cls_alias = clean_id(type_name(spec.origin_type))
            method_name_alias = f"{cls_alias}_{method_name}"
            spec.builder.ensure_object_imported(
                getattr(spec.attrs, method_name), method_name_alias
            )
            method_args = spec.expression
            return f"{method_name_alias}({method_args})"


@register
def pack_final(spec: ValueSpec) -> Optional[Expression]:
    if is_final(spec.type):
        return PackerRegistry.get(spec.copy(type=get_args(spec.type)[0]))


@register
def pack_any(spec: ValueSpec) -> Optional[Expression]:
    if spec.type is Any:
        return spec.expression


def pack_union(
    spec: ValueSpec, args: tuple[type, ...], prefix: str = "union"
) -> Expression:
    if spec.type is spec.owner and spec.field_ctx.packer:
        return spec.field_ctx.packer
    lines = CodeLines()

    method_name = (
        f"__pack_{prefix}_{spec.builder.cls.__name__}_{spec.field_ctx.name}__"
        f"{random_hex()}"
    )

    if not spec.field_ctx.packer:
        method_args = ", ".join(
            filter(None, ("value", spec.builder.get_pack_method_flags()))
        )
        if spec.builder.is_nailed:
            union_packer = (
                f"{spec.self_attrs_name}.{method_name}({method_args})"
            )
        else:
            union_packer = f"{method_name}({method_args})"
        spec.field_ctx.packer = union_packer

    method_args = "self, value" if spec.builder.is_nailed else "value"
    default_kwargs = spec.builder.get_pack_method_default_flag_values()
    if default_kwargs:
        lines.append(f"def {method_name}({method_args}, {default_kwargs}):")
    else:
        lines.append(f"def {method_name}({method_args}):")
    packers: list[str] = []
    packer_arg_types: dict[str, list[type]] = {}
    for type_arg in args:
        packer = PackerRegistry.get(
            spec.copy(type=type_arg, expression="value", owner=spec.type)
        )
        if packer not in packers:
            if packer == "value":
                packers.insert(0, packer)
            else:
                packers.append(packer)
        packer_arg_types.setdefault(packer, []).append(type_arg)

    if len(packers) == 1 and packers[0] == "value":
        return spec.expression

    with lines.indent():
        for packer in packers:
            packer_arg_type_names = []
            for packer_arg_type in packer_arg_types[packer]:
                if is_generic(packer_arg_type):
                    packer_arg_type = get_type_origin(packer_arg_type)
                packer_arg_type_name = clean_id(type_name(packer_arg_type))
                spec.builder.ensure_object_imported(
                    packer_arg_type, packer_arg_type_name
                )
                if packer_arg_type_name not in packer_arg_type_names:
                    packer_arg_type_names.append(packer_arg_type_name)
            if len(packer_arg_type_names) > 1:
                packer_arg_type_check = (
                    f"in ({', '.join(packer_arg_type_names)})"
                )
            else:
                packer_arg_type_check = f"is {packer_arg_type_names[0]}"
            if packer == "value":
                with lines.indent(
                    f"if value.__class__ {packer_arg_type_check}:"
                ):
                    lines.append(f"return {packer}")
            else:
                with lines.indent("try:"):
                    lines.append(f"return {packer}")
                with lines.indent("except Exception:"):
                    lines.append("pass")
        field_type = spec.builder.get_type_name_identifier(
            typ=spec.type,
            resolved_type_params=spec.builder.get_field_resolved_type_params(
                spec.field_ctx.name
            ),
        )
        if spec.builder.is_nailed:
            lines.append(
                "raise InvalidFieldValue("
                f"'{spec.field_ctx.name}',{field_type},value,type(self))"
            )
        else:
            lines.append("raise ValueError(value)")
    lines.append(
        f"setattr({spec.cls_attrs_name}, '{method_name}', {method_name})"
    )
    if spec.builder.get_config().debug:
        print(f"{type_name(spec.builder.cls)}:")
        print(lines.as_text())

    exec(lines.as_text(), spec.builder.globals, spec.builder.__dict__)

    method_args = ", ".join(
        filter(None, (spec.expression, spec.builder.get_pack_method_flags()))
    )
    if spec.builder.is_nailed:
        return f"{spec.self_attrs_name}.{method_name}({method_args})"
    else:
        spec.builder.ensure_object_imported(
            getattr(spec.attrs, method_name), method_name
        )
        return f"{method_name}({method_args})"


def pack_literal(spec: ValueSpec) -> Expression:
    spec.builder.add_type_modules(spec.type)
    lines = CodeLines()
    method_name = (
        f"__pack_literal_{spec.builder.cls.__name__}_{spec.field_ctx.name}__"
        f"{random_hex()}"
    )
    method_args = "self, value" if spec.builder.is_nailed else "value"
    default_kwargs = spec.builder.get_pack_method_default_flag_values()
    if default_kwargs:
        lines.append(f"def {method_name}({method_args}, {default_kwargs}):")
    else:
        lines.append(f"def {method_name}({method_args}):")
    resolved_type_params = spec.builder.get_field_resolved_type_params(
        spec.field_ctx.name
    )
    with lines.indent():
        for literal_value in get_literal_values(spec.type):
            value_type = type(literal_value)
            packer = PackerRegistry.get(
                spec.copy(type=value_type, expression="value")
            )
            if isinstance(literal_value, enum.Enum):
                enum_type_name = spec.builder.get_type_name_identifier(
                    typ=value_type,
                    resolved_type_params=resolved_type_params,
                )
                with lines.indent(
                    f"if value == {enum_type_name}.{literal_value.name}:"
                ):
                    lines.append(f"return {packer}")
            elif isinstance(
                literal_value,
                (int, str, bytes, bool, NoneType),  # type: ignore
            ):
                with lines.indent(f"if value == {literal_value!r}:"):
                    lines.append(f"return {packer}")
        field_type = spec.builder.get_type_name_identifier(
            typ=spec.type,
            resolved_type_params=resolved_type_params,
        )
        if spec.builder.is_nailed:
            lines.append(
                f"raise InvalidFieldValue('{spec.field_ctx.name}',"
                f"{field_type},value,type(self))"
            )
        else:
            lines.append("raise ValueError(value)")
    lines.append(
        f"setattr({spec.cls_attrs_name}, '{method_name}', {method_name})"
    )
    if spec.builder.get_config().debug:
        print(f"{type_name(spec.builder.cls)}:")
        print(lines.as_text())
    exec(lines.as_text(), spec.builder.globals, spec.builder.__dict__)
    method_args = ", ".join(
        filter(None, (spec.expression, spec.builder.get_pack_method_flags()))
    )
    return f"{spec.self_attrs_name}.{method_name}({method_args})"


@register
def pack_special_typing_primitive(spec: ValueSpec) -> Optional[Expression]:
    if is_special_typing_primitive(spec.origin_type):
        if is_union(spec.type):
            resolved_type_params = spec.builder.get_field_resolved_type_params(
                spec.field_ctx.name
            )
            if is_optional(spec.type, resolved_type_params):
                arg = not_none_type_arg(
                    get_args(spec.type), resolved_type_params
                )
                pv = PackerRegistry.get(spec.copy(type=arg))
                return expr_or_maybe_none(spec, pv)
            else:
                return pack_union(spec, get_args(spec.type))
        elif spec.origin_type is typing.AnyStr:
            raise UnserializableDataError(
                "AnyStr is not supported by mashumaro"
            )
        elif is_type_var_any(spec.type):
            return spec.expression
        elif is_type_var(spec.type):
            constraints = getattr(spec.type, "__constraints__")
            if constraints:
                return pack_union(spec, constraints, "type_var")
            else:
                if type_var_has_default(spec.type):
                    bound = get_type_var_default(spec.type)
                else:
                    bound = getattr(spec.type, "__bound__")
                # act as if it was Optional[bound]
                pv = PackerRegistry.get(spec.copy(type=bound))
                return expr_or_maybe_none(spec, pv)
        elif is_new_type(spec.type):
            return PackerRegistry.get(spec.copy(type=spec.type.__supertype__))
        elif is_literal(spec.type):
            return pack_literal(spec)
        elif spec.type is typing_extensions.LiteralString:
            return PackerRegistry.get(spec.copy(type=str))
        elif is_self(spec.type):
            method_name = spec.builder.get_pack_method_name(
                format_name=spec.builder.format_name
            )
            method_loc = (
                spec.builder.cls if spec.builder.is_nailed else spec.attrs
            )
            if (
                get_class_that_defines_method(method_name, method_loc)
                != method_loc
                # not hasattr(self.cls, method_name)
                and spec.builder.get_pack_method_name(
                    format_name=spec.builder.format_name,
                    encoder=spec.builder.encoder,
                )
                != method_name
            ):
                builder = spec.builder.__class__(
                    spec.builder.cls,
                    dialect=spec.builder.dialect,
                    format_name=spec.builder.format_name,
                    default_dialect=spec.builder.default_dialect,
                    attrs=method_loc,
                    attrs_registry=(
                        spec.attrs_registry
                        if not spec.builder.is_nailed
                        else None
                    ),
                )
                builder.add_pack_method()
            flags = spec.builder.get_pack_method_flags(spec.builder.cls)
            if spec.builder.is_nailed:
                return f"{spec.expression}.{method_name}({flags})"
            else:
                method_args = spec.expression
                return f"_cls.{method_name}({method_args})"
        elif is_required(spec.type) or is_not_required(spec.type):
            return PackerRegistry.get(spec.copy(type=get_args(spec.type)[0]))
        elif is_unpack(spec.type):
            packer = PackerRegistry.get(spec.copy(type=get_args(spec.type)[0]))
            return f"*{packer}"
        elif is_type_var_tuple(spec.type):
            return PackerRegistry.get(spec.copy(type=tuple[Any, ...]))
        elif isinstance(spec.type, ForwardRef):
            evaluated = evaluate_forward_ref(spec.type)
            if evaluated is not None:
                return PackerRegistry.get(spec.copy(type=evaluated))
        elif is_type_alias_type(spec.type):
            return PackerRegistry.get(spec.copy(type=spec.type.__value__))
        elif is_readonly(spec.type):
            return PackerRegistry.get(spec.copy(type=get_args(spec.type)[0]))
        raise UnserializableDataError(
            f"{spec.type} as a field type is not supported by mashumaro"
        )


@register
def pack_number_and_bool_and_none(spec: ValueSpec) -> Optional[Expression]:
    if spec.origin_type in (int, float, bool, NoneType, None):
        return spec.expression


@register
def pack_date_objects(spec: ValueSpec) -> Optional[Expression]:
    if spec.origin_type in (datetime.datetime, datetime.date, datetime.time):
        return f"{spec.expression}.isoformat()"


@register
def pack_timedelta(spec: ValueSpec) -> Optional[Expression]:
    if spec.origin_type is datetime.timedelta:
        return f"{spec.expression}.total_seconds()"


@register
def pack_timezone(spec: ValueSpec) -> Optional[Expression]:
    if spec.origin_type is datetime.timezone:
        return f"{spec.expression}.tzname(None)"


@register
def pack_zone_info(spec: ValueSpec) -> Optional[Expression]:
    if spec.origin_type is zoneinfo.ZoneInfo:
        return f"str({spec.expression})"


@register
def pack_uuid(spec: ValueSpec) -> Optional[Expression]:
    if spec.origin_type is uuid.UUID:
        return f"str({spec.expression})"


@register
def pack_ipaddress(spec: ValueSpec) -> Optional[Expression]:
    if spec.origin_type in (
        ipaddress.IPv4Address,
        ipaddress.IPv6Address,
        ipaddress.IPv4Network,
        ipaddress.IPv6Network,
        ipaddress.IPv4Interface,
        ipaddress.IPv6Interface,
    ):
        return f"str({spec.expression})"


@register
def pack_decimal(spec: ValueSpec) -> Optional[Expression]:
    if spec.origin_type is Decimal:
        return f"str({spec.expression})"


@register
def pack_fraction(spec: ValueSpec) -> Optional[Expression]:
    if spec.origin_type is Fraction:
        return f"str({spec.expression})"


def pack_tuple(spec: ValueSpec, args: tuple[type, ...]) -> Expression:
    if not args:
        if spec.type in (Tuple, tuple):
            args = [Any, ...]  # type: ignore
        else:
            return "[]"
    elif len(args) == 1 and args[0] == ():
        if not PY_311_MIN:
            return "[]"
    if len(args) == 2 and args[1] is Ellipsis:
        packer = PackerRegistry.get(
            spec.copy(type=args[0], expression="value", could_be_none=True)
        )
        return f"[{packer} for value in {spec.expression}]"
    else:
        arg_indexes: list[Union[int, tuple[int, Union[int, None]]]] = []
        unpack_idx: Optional[int] = None
        for arg_idx, type_arg in enumerate(args):
            if is_unpack(type_arg):
                if unpack_idx is not None:
                    raise TypeError(
                        "Multiple unpacks are disallowed within a single type "
                        f"parameter list for {type_name(spec.type)}"
                    )
                unpack_idx = arg_idx
                if len(args) == 1:
                    arg_indexes.append((arg_idx, None))
                elif arg_idx < len(args) - 1:
                    arg_indexes.append((arg_idx, arg_idx + 1 - len(args)))
                else:
                    arg_indexes.append((arg_idx, None))
            else:
                if unpack_idx is None:
                    arg_indexes.append(arg_idx)
                else:
                    arg_indexes.append(arg_idx - len(args))
        packers: list[Expression] = []
        for _idx, _arg_idx in enumerate(arg_indexes):
            if isinstance(_arg_idx, tuple):
                p_expr = f"{spec.expression}[{_arg_idx[0]}:{_arg_idx[1]}]"
            else:
                p_expr = f"{spec.expression}[{_arg_idx}]"
            packer = PackerRegistry.get(
                spec.copy(
                    type=args[_idx],
                    expression=p_expr,
                    could_be_none=True,
                )
            )
            if packer != "*[]":
                packers.append(packer)
        return f"[{', '.join(packers)}]"


def pack_named_tuple(spec: ValueSpec) -> Expression:
    resolved = resolve_type_params(spec.origin_type, get_args(spec.type))[
        spec.origin_type
    ]
    annotations = {
        k: resolved.get(v, v)
        for k, v in get_annotations(spec.origin_type, eval_str=True).items()
    }
    fields = getattr(spec.type, "_fields", ())
    packers = []
    as_dict = spec.builder.get_dialect_or_config_option(
        "namedtuple_as_dict", False
    )
    serialize_option = get_overridden_serialization_method(spec)
    if serialize_option is not None:
        if serialize_option == "as_dict":
            as_dict = True
        elif serialize_option == "as_list":
            as_dict = False
        else:
            raise UnsupportedSerializationEngine(
                field_name=spec.field_ctx.name,
                field_type=spec.type,
                holder_class=spec.builder.cls,
                engine=serialize_option,
            )
    for idx, field in enumerate(fields):
        packer = PackerRegistry.get(
            spec.copy(
                type=annotations.get(field, Any),
                expression=f"{spec.expression}[{idx}]",
                could_be_none=True,
            )
        )
        packers.append(packer)
    if as_dict:
        kv = (f"'{key}': {value}" for key, value in zip(fields, packers))
        return f"{{{', '.join(kv)}}}"
    else:
        return f"[{', '.join(packers)}]"


def pack_typed_dict(spec: ValueSpec) -> Expression:
    resolved = resolve_type_params(spec.origin_type, get_args(spec.type))[
        spec.origin_type
    ]
    annotations = {
        k: resolved.get(v, v)
        for k, v in get_annotations(spec.origin_type, eval_str=True).items()
    }
    all_keys = list(annotations.keys())
    required_keys = set(getattr(spec.type, "__required_keys__", all_keys))
    optional_keys = set(getattr(spec.type, "__optional_keys__", []))

    # workaround for https://github.com/python/cpython/issues/97727
    for key, annotation in annotations.items():
        if isinstance(annotation, ForwardRef):
            annotation = evaluate_forward_ref(annotation)
            if get_type_origin(annotation) is NotRequired:
                required_keys.discard(key)
                optional_keys.add(key)

    lines = CodeLines()
    method_name = (
        f"__pack_typed_dict_{spec.builder.cls.__name__}_"
        f"{spec.field_ctx.name}__{random_hex()}"
    )
    method_args = "self, value" if spec.builder.is_nailed else "value"
    default_kwargs = spec.builder.get_pack_method_default_flag_values()
    if default_kwargs:
        lines.append(f"def {method_name}({method_args}, {default_kwargs}):")
    else:
        lines.append(f"def {method_name}({method_args}):")
    with lines.indent():
        lines.append("d = {}")
        for key in sorted(required_keys, key=all_keys.index):
            packer = PackerRegistry.get(
                spec.copy(
                    type=annotations[key],
                    expression=f"value['{key}']",
                    could_be_none=True,
                    owner=spec.type,
                )
            )
            lines.append(f"d['{key}'] = {packer}")
        for key in sorted(optional_keys, key=all_keys.index):
            lines.append(f"key_value = value.get('{key}', MISSING)")
            with lines.indent("if key_value is not MISSING:"):
                packer = PackerRegistry.get(
                    spec.copy(
                        type=annotations[key],
                        expression="key_value",
                        could_be_none=True,
                        owner=spec.type,
                    )
                )
                lines.append(f"d['{key}'] = {packer}")
        lines.append("return d")
    lines.append(
        f"setattr({spec.cls_attrs_name}, '{method_name}', {method_name})"
    )
    if spec.builder.get_config().debug:
        print(f"{type_name(spec.builder.cls)}:")
        print(lines.as_text())
    exec(lines.as_text(), spec.builder.globals, spec.builder.__dict__)
    method_args = ", ".join(
        filter(None, (spec.expression, spec.builder.get_pack_method_flags()))
    )
    return f"{spec.self_attrs_name}.{method_name}({method_args})"


@register
def pack_collection(spec: ValueSpec) -> Optional[Expression]:
    if not issubclass(spec.origin_type, Collection):
        return None
    elif issubclass(spec.origin_type, enum.Enum):
        return None

    args = get_args(spec.type)

    def inner_expr(
        arg_num: int = 0, v_name: str = "value", v_type: Optional[type] = None
    ) -> Expression:
        if v_type:
            return PackerRegistry.get(
                spec.copy(type=v_type, expression=v_name)
            )
        else:
            if args and len(args) > arg_num:
                type_arg: Any = args[arg_num]
            else:
                type_arg = Any
            return PackerRegistry.get(
                spec.copy(
                    type=type_arg,
                    expression=v_name,
                    could_be_none=True,
                    field_ctx=spec.field_ctx.copy(metadata={}),
                )
            )

    def _make_sequence_expression(ie: Expression) -> Expression:
        if ie == "value":
            if spec.origin_type in spec.no_copy_collections:
                return spec.expression
            elif spec.origin_type is list:
                return f"{spec.expression}.copy()"
        return f"[{ie} for value in {spec.expression}]"

    def _make_mapping_expression(ke: Expression, ve: Expression) -> Expression:
        if ke == "key" and ve == "value":
            if spec.origin_type in spec.no_copy_collections:
                return spec.expression
            elif spec.origin_type is dict:
                return f"{spec.expression}.copy()"
        return f"{{{ke}: {ve} for key, value in {spec.expression}.items()}}"

    if issubclass(spec.origin_type, typing.ByteString):  # type: ignore
        spec.builder.ensure_object_imported(encodebytes)
        return f"encodebytes({spec.expression}).decode()"
    elif issubclass(spec.origin_type, str):
        return spec.expression
    elif issubclass(spec.origin_type, tuple):
        if is_named_tuple(spec.origin_type):
            return pack_named_tuple(spec)
        elif ensure_generic_collection(spec):
            return pack_tuple(spec, args)
    elif ensure_generic_collection_subclass(spec, list, deque, Set):
        ie = inner_expr()
        return _make_sequence_expression(ie)
    elif ensure_generic_mapping(spec, args, ChainMap):
        ke = inner_expr(0, "key")
        ve = inner_expr(1)
        return (
            f"[{{{ke}: {ve} for key, value in m.items()}} "
            f"for m in {spec.expression}.maps]"
        )
    elif ensure_generic_mapping(spec, args, OrderedDict):
        ke = inner_expr(0, "key")
        ve = inner_expr(1)
        return _make_mapping_expression(ke, ve)
    elif ensure_generic_mapping(spec, args, Counter):
        ke = inner_expr(0, "key")
        ve = inner_expr(1, v_type=int)
        return _make_mapping_expression(ke, ve)
    elif is_typed_dict(spec.origin_type):
        return pack_typed_dict(spec)
    elif ensure_generic_mapping(spec, args, Mapping):
        ke = inner_expr(0, "key")
        ve = inner_expr(1)
        return _make_mapping_expression(ke, ve)
    elif ensure_generic_collection_subclass(spec, Sequence):
        ie = inner_expr()
        return _make_sequence_expression(ie)


@register
def pack_pathlike(spec: ValueSpec) -> Optional[Expression]:
    if issubclass(spec.origin_type, os.PathLike):
        return f"{spec.expression}.__fspath__()"


@register
def pack_enum(spec: ValueSpec) -> Optional[Expression]:
    if issubclass(spec.origin_type, enum.Enum):
        return f"{spec.expression}.value"


@register
def pack_pattern(spec: ValueSpec) -> Optional[Expression]:
    if spec.origin_type in (typing.Pattern, re.Pattern):
        return f"{spec.expression}.pattern"



from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List

import msgpack

from mashumaro import DataClassDictMixin
from mashumaro.config import ADD_DIALECT_SUPPORT, BaseConfig
from mashumaro.dialect import Dialect
from mashumaro.mixins.msgpack import DataClassMessagePackMixin


class MyDialect(Dialect):
    serialization_strategy = {
        bytes: {
            "serialize": lambda x: x.decode(),
            "deserialize": lambda x: x.encode(),
        }
    }


@dataclass
class InnerDataClass(DataClassDictMixin):
    x: bytes


@dataclass
class DataClass(DataClassMessagePackMixin):
    x: bytes
    inner: InnerDataClass


def test_data_class_msgpack_mixin():
    assert DataClassMessagePackMixin.from_msgpack(b"") is None
    assert DataClassMessagePackMixin().to_msgpack() is None


def test_to_msgpack():
    @dataclass
    class DataClass(DataClassMessagePackMixin):
        x: List[int]

    dumped = msgpack.packb({"x": [1, 2, 3]})
    assert DataClass([1, 2, 3]).to_msgpack() == dumped


def test_from_msgpack():
    @dataclass
    class DataClass(DataClassMessagePackMixin):
        x: List[int]

    dumped = msgpack.packb({"x": [1, 2, 3]})
    assert DataClass.from_msgpack(dumped) == DataClass([1, 2, 3])


def test_to_msg_pack_datetime():
    @dataclass
    class DataClass(DataClassMessagePackMixin):
        x: datetime

    dt = datetime(2018, 10, 29, 12, 46, 55, 308495)
    dumped = msgpack.packb({"x": dt.isoformat()})
    assert DataClass(dt).to_msgpack() == dumped


def test_msgpack_with_bytes():
    @dataclass
    class DataClass(DataClassMessagePackMixin):
        x: bytes
        y: bytearray

    instance = DataClass(b"123", bytearray(b"456"))
    dumped = msgpack.packb({"x": b"123", "y": bytearray(b"456")})
    assert DataClass.from_msgpack(dumped) == instance
    assert instance.to_msgpack() == dumped


def test_msgpack_with_serialization_strategy():
    @dataclass
    class DataClass(DataClassMessagePackMixin):
        x: bytes

        class Config(BaseConfig):
            serialization_strategy = {
                bytes: {
                    "serialize": lambda x: x.decode(),
                    "deserialize": lambda x: x.encode(),
                }
            }

    instance = DataClass(b"123")
    dumped = msgpack.packb({"x": "123"})
    assert DataClass.from_dict({"x": "123"}) == instance
    assert DataClass.from_msgpack(dumped) == instance
    assert instance.to_dict() == {"x": "123"}
    assert instance.to_msgpack() == dumped


def test_msgpack_with_dialect():
    @dataclass
    class DataClass(DataClassMessagePackMixin):
        x: bytes

        class Config(BaseConfig):
            dialect = MyDialect

    instance = DataClass(b"123")
    dumped_dialect = msgpack.packb({"x": "123"})
    assert DataClass.from_dict({"x": "123"}) == instance
    assert DataClass.from_msgpack(dumped_dialect) == instance
    assert instance.to_dict() == {"x": "123"}
    assert instance.to_msgpack() == dumped_dialect


def test_msgpack_with_dialect_support():
    @dataclass
    class DataClass(DataClassMessagePackMixin):
        x: bytes

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]

    instance = DataClass(b"123")
    dumped = msgpack.packb({"x": b"123"})
    dumped_dialect = msgpack.packb({"x": "123"})
    assert DataClass.from_dict({"x": "MTIz\n"}) == instance
    assert DataClass.from_dict({"x": "123"}, dialect=MyDialect) == instance
    assert DataClass.from_msgpack(dumped) == instance
    assert (
        DataClass.from_msgpack(dumped_dialect, dialect=MyDialect) == instance
    )
    assert instance.to_dict() == {"x": "MTIz\n"}
    assert instance.to_dict(dialect=MyDialect) == {"x": "123"}
    assert instance.to_msgpack() == dumped
    assert instance.to_msgpack(dialect=MyDialect) == dumped_dialect


def test_msgpack_with_custom_encoder_and_decoder():
    def decoder(data) -> Dict[str, bytes]:
        def to_lower(d):
            result = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    result[k] = to_lower(v)
                else:
                    result[k] = v.lower()
            return result

        return to_lower(msgpack.loads(data))

    def encoder(data: Dict[str, bytes]) -> bytes:
        def to_upper(d):
            result = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    result[k] = to_upper(v)
                else:
                    result[k] = v.upper()
            return result

        return msgpack.dumps(to_upper(data))

    instance = DataClass(b"abc", InnerDataClass(b"def"))
    dumped = msgpack.packb({"x": b"ABC", "inner": {"x": b"DEF"}})
    assert instance.to_msgpack(encoder=encoder) == dumped
    assert DataClass.from_msgpack(dumped, decoder=decoder) == instance
